;;  datoura.lsp. The Data Touring Machine. Copyright (C) 2004 Steve Eichblatt
;;
;;  This program is free software; you can redistribute it and/or modify
;;  it under the terms of the GNU General Public License as published by
;;  the Free Software Foundation; either version 2 of the License, or
;;  (at your option) any later version.
;;
;;  This program is distributed in the hope that it will be useful,
;;  but WITHOUT ANY WARRANTY; without even the implied warranty of
;;  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;  GNU General Public License for more details.
;;
;;  You should have received a copy of the GNU General Public License
;;  along with this program; if not, write to the Free Software
;;  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
;;
;;  Contact: Steve Eichblatt
;;  620 E San Salvador San Jose, CA 95112. steve@eichblatt.us

;; -------------------------------------------------------- session-id
(defun datoura-session-id ()
"session-id () -- returns a string that identifies a datoura session
based on the date and time. Should be unique, but this isn't guaranteed."
  (multiple-value-bind (sec min hour day month year weekday dls tzone)
                         (get-decoded-time)
      (format nil "dtr~4D~2@{~2,'0D~}_~2@{~2,'0D~}~2,'0D" 
         year month day hour min sec)))

(unless (ignore-errors *datoura-session-id*) 
        (defparameter *datoura-session-id* (datoura-session-id)))
(export '*datoura-session-id*)

(load (compile-file "./utils"))
(load (compile-file "./graphics"))
(load (compile-file "./color"))
(load (compile-file "./plot"))
(load (compile-file "./csv-parser"))
(load (compile-file "./dtable"))
(defpackage :datoura
           (:nicknames :datoura)
           (:use :dtable :plot :color :graphics :common-lisp :utils
                 :common-lisp-user )
           (:export #:hello #:bye #:*session-id* #:*base-path*))
(in-package :datoura)

(graphics::set-graphics-path
  #-(or gcl win32)
    (ensure-directories-exist 
      (pathname (concatenate 'string (namestring (user-homedir-pathname))
      ".datoura-sessions/" *datoura-session-id* "/graph/")))
  #+(or gcl win32)
      (user-homedir-pathname)) 

(setf hello "hello")
(defmacro bye  () 
  (cleanup) `(in-package :common-lisp-user))
(defmacro exit () `(in-package :common-lisp-user))
;(memoize 'dt-project :key #'identity :test #'equal)

(defparameter *base-path* 
#-gcl
  (ensure-directories-exist 
    (pathname (concatenate 'string (namestring (user-homedir-pathname))
    ".datoura-sessions/" *datoura-session-id* "/")))
#+gcl
  (user-homedir-pathname)
)

;; ------------------------------------------------------------ cleanup
(defun cleanup ()
"cleanup () -- clean up the files created by this session..."
#+(and clisp unix)
  (ext:shell (format nil "rm -rf ~A" *base-path*)))
;; ------------------------------------------------------------ plot-cdf
(defun plot-cdf (dt expr &key by where (colors (cdr color:colorlist))
                   (symbol :mdsqe) join options) 
"plot-cdr (dt expr &key by where (colors (cdr color:colorlist))
                   options) -- hist+cdf plot (by byvar).  "
  (let ((data (dt-project dt expr :by by :where where))
        (gl (make-glist)) (ic -1))
    (unless (assoc-val 'xlabel options) 
      (setf options (acons 'xlabel expr options)))
;    (unless (assoc-val 'ylabel options) 
;      (setf options (acons 'ylabel (second expr) options)))
    (unless (assoc-val 'title options) 
      (setf options (acons 'title (format nil "by ~A" by) options)))
    (dbg :plot-cdf "hash-keys ~A" (keys data))
    (maphash #'(lambda (k v) 
      (let* ((stat (simple-stats v))
             (mean (assoc-val 'mean stat)) (rms (assoc-val 'rms stat))
             (stderr (* 2 (/ rms (sqrt (assoc-val 'N stat))))))
        (glist-add gl 
          (histogram v :normalize 100
                   :dash t
                   :options (acons 'yaxis :y0 options)
                   :color (elt colors (incf ic))) 
           (when (> (length v) 10000)    ;; this should be a function in utils
              (let* ((lv (length v)) (stride (floor (/ lv 5000))))
                (setf v (concatenate 'vector (subseq v 0 10) 
                           (extract v (range 10 (- lv 10) stride))
                           (subseq v (- lv 10) lv)))))
          (scatter-plot (coerce (cdf v :normalize 100) 'list)  ; dies if list 
                   :name k :join join :symbol symbol           ; is too long.
                   :options (acons 'yaxis :y1 options)
                   :color (elt colors ic))
          (set-color (elt colors ic))
          (diamond `((,mean .18) (,stderr .05)) '(axes . (:x0 :ry)))
          (unset-color))))
    data)
    gl))
;; ------------------------------------------------------------ scatter-by
(defun scatter-by (dt expr by &key where (colors (cdr color:colorlist))
                   (symbol :mdsqe) options) 
"scatter-by (dt expr by &key where (colors (cdr color:colorlist))
                   (symbol :mdsqe) options) -- scatter plot by byvar.  "
  (let ((data (dt-project-cols dt expr :by by :where where))
        (gl (make-glist)) (ic -1))
    (unless (assoc-val 'xlabel options) 
      (setf options (acons 'xlabel (first  expr) options)))
    (unless (assoc-val 'ylabel options) 
      (setf options (acons 'ylabel (second expr) options)))
    (unless (assoc-val 'title options) 
      (setf options (acons 'title (format nil "by ~A" by) options)))
    (dbg :scatter-by "hash-keys ~A" (keys (first data)))
    (maphash #'(lambda (k v) 
      (glist-add gl (scatter-plot (thread v (gethash k (second data))) 
                 :name k :symbol symbol :profile t :options options 
                 :color (elt colors (incf ic))))) 
    (first data))
    gl))

;; this would be an example to print the values of "w" in the "w" groups:
;(maphash #'(lambda (key val) (print (extract (gethash "w" (dtable-data dt)) val))) byw)
;; This takes statistics (N, mean, sigma) of the column "birthyear" for 
;; the groups...
;; it ought to be a command like:
;; (stats '("birthyear") :by '("birthyear" "w"))
;;(maphash #'(lambda (key val) (print (simple-stats (extract (gethash "birthyear" (dtable-data dt)) val)))) byr)
;; or better yet (stats '(birthyear player_id) :where #'cutf :by '("w"))
;;
;; An example of a pretty complex (where-by):
;;
;;(setf fcol    
;;   (formula-col dt #'(lambda (x y) (and x y (mod x y))) "w" "birthyear"))
;;(setf where-fg250 (positions #'(lambda (x) (> x 250)) fcol))
;;(maphash 
;;   #'(lambda (key val) 
;;             (format t "~A -> ~A~%" key (intersection where-fg250 val))) 
;;   (gethash '("birthyear") (dtable-groups dt)))
;; note- this only prints the result, it doesn't save or return it.
;; it will have to create a hash to return it. 
