;;  dtable.lsp. The Data Touring Machine. Copyright (C) 2004 Steve Eichblatt
;;
;;  This program is free software; you can redistribute it and/or modify
;;  it under the terms of the GNU General Public License as published by
;;  the Free Software Foundation; either version 2 of the License, or
;;  (at your option) any later version.
;;
;;  This program is distributed in the hope that it will be useful,
;;  but WITHOUT ANY WARRANTY; without even the implied warranty of
;;  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;  GNU General Public License for more details.
;;
;;  You should have received a copy of the GNU General Public License
;;  along with this program; if not, write to the Free Software
;;  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
;;
;;  Contact: Steve Eichblatt
;;  620 E San Salvador San Jose, CA 95112. steve@eichblatt.us

(unless (find-package :utils) (load (compile-file "utils")))
(unless (find-package :csv-parser) (load (compile-file "csv-parser")))
(defpackage :dtable
           (:use :common-lisp :csv-parser :utils)
           (:export #:dtable-export
                    #:dtable-import #:dt-drop #:dt-concat #:dt-random-subset
                    #:dtable-header #:dtable-label #:dtable-label 
                    #:dt-project #:dt-subset #:dt-join #:dt-split
                    #:dtable-types #:dtable-data #:dtable-ordering
                    #:dtable-groups #:dtable-properties 
                    #:group #:print-dtable #:dt-where
                    #:bygroups #:groupkeys #:dtable-sort))
(in-package :dtable)
(defconstant *dtable-print-length* 5 
"how many elements of the columns to print")

;; --------------------------------------------------------- dtable
(defstruct (dtable (:print-function print-dtable))
  "dtable == The data table structure, containing the data, headers, groups
   and state of the data table"
  header            ; symbol names of columns
  label             ; string names of columns
 (data   (make-hash-table :test #'equal)) ; hash of columns
 (types  (make-hash-table :test #'equal)) ; hash of column types
 (groups (make-hash-table :test #'equal)) ; hash of most recent groups
  hidden            ;
  active            ;
  ordering          ; current ordering
  deleted           ; vector of observations to ignore forever.
  properties        ; general purpose
)

;; this stuff belongs in the csv-parser file
;;===================================================================
;; new approach -- use the csv-parser built-in macros to load data:
;; --------------------------------------------------------- convert-to-num
(defun convert-to-num (seq)
"convert-to-num (seq) -- convert the elements in sequence to numbers iff all
of them are numbers"
  (if (stringp (elt-0 seq))
  (let ((accum (make-array (length seq) :adjustable t :fill-pointer 0)) 
        (counter 0))
    (loop 
      (let ((elem (elt seq counter)))
        (dbg :convert "counter: ~A" counter)
        (vector-push-extend 
          (handler-case 
            (cond ((= (length elem) 0) () )
                  ((equalp "." elem) () )
                  ((equalp "()" elem) () )
                  (t (read-from-string elem)))
          (stream-error () elem)) accum)
        (unless (or (null (elt accum counter))
           (numberp (elt accum counter))) (return seq))
        (when (>= (incf counter) (length seq)) (return accum)))))
   seq))
;; --------------------------------------------------------- parse-csv-line
(defun parse-csv-line (x)
  "parse-csv-line (x) -- given a line from read-csv-line reads and
  casts values to their appropriate type.  "
  (let ((result ()) (elem ()))
  (dolist (elemarray x)
    (setq elem (coerce elemarray 'string))
    (push (handler-case 
      (cond ((= (length elem) 0) () )
            ((equalp "." elem) () )
            ((equalp "()" elem) () )
            ((equalp 'symbol (type-of (read-from-string elem))) elem)
            (t (read-from-string elem)))
      (stream-error () elem))  ;; handles weird cases like "8E12345" -> string
    result))
    (nreverse result)))
;; -------------------------------------------------- csv-to-lines
(defun csv-to-lines (filename)
  (let ((lines ()))
      (do-csv-file ((line fh) filename) 
        (identity fh)
        (push line lines))
;;(        (push (parse-csv-line line) lines))
      (nreverse lines)))             ;; can't i avoid some reversing?
;; -------------------------------------------------- string-trim-std
(defun string-trim-std (s)
  (string-trim '(#\Space #\Tab #\Newline #\") s)) ;"
;; -------------------------------------------------- csv-to-lines
(defun csv-to-lines2 (filename)
(with-open-file (ifile filename :direction :input)
  (let ((iline 0) (line nil) (flds ())
        (lines (make-array 10000 :element-type 'string 
                  :adjustable t :fill-pointer 0))
        (header (read-csv-line ifile)))
  (vector-push-extend header lines)
  (dbg :import "~&Length Header ~D~%" (length header))
  (loop (setf line (read-line ifile nil 'eof))
        (if (equal line 'eof) (return)) 
        (incf iline)
        (setf flds (mapcar #'string-trim-std (split-string line #'commap)))
        (unless (= (length flds) (length header)) 
           (format t "~&More fields than header line ~D ~%~A~% ~A ~%" 
                     iline line (length flds)) 
           (setf flds 
             (with-input-from-string (sss line) (read-csv-line sss)))
           (format t "~&correct line ~A ~%" flds))
        (vector-push-extend flds lines))
  lines)))
;;===================================================================
;; --------------------------------------------------- lines-to-hash
(defun lines-to-hash (lines)
  "lines-to-hash (lines)
  given the list of lines (as from csv-to-lines) stores the data 
  in a hash table, which it returns, along with the header.  "
  (let* ((label (elt lines 0)) (line nil) 
         (dataHash (make-hash-table)) 
         (hs (mapcar #'(lambda (x) 
                          (substitute #\_ #\SPACE x)) label))
         (header (mapcar #'read-from-string hs)))

      (dbg :load "header ~A" header)
  ;; load the data into hashes, with key based on the header line.
  ;; problem: It should be done directly while reading the file to save time.
  ;; But, it does work!
    ; signal error if 
    ; (> (length (header)) (length (remove-duplicates (header))))

    (dolist (parm header)
      (setf (gethash parm dataHash) (make-array (length lines) 
                                     :adjustable t :fill-pointer 0)))
;    (setf (gethash ':n dataHash) 
;          (make-array (length lines) :adjustable t :fill-pointer 0))

    (let ((n 0))
      (dotimes (j (- (length lines) 1))
;        (vector-push n (gethash ':n dataHash))
        (setf line (elt lines (+ j 1)))
	(dotimes (k (length header))
	  (vector-push (ignore-errors (elt line k)) 
               (gethash (elt header k) dataHash)))
        (incf n)))

    (dbg :load "dataHash ~A" dataHash)
    (values dataHash label)))
;; --------------------------------------------------- csv-to-hash
(defun csv-to-hash (filename)
  (lines-to-hash (csv-to-lines2 filename)))
;; --------------------------------------------------- add-col
(defun add-col (dt colname coldata)
"add-col (dt colname coltype coldata) -- add the column to dtable dt.
  PROBLEMS: 
  --  Ensure that colname is unique, modify if needed. Return colname.
  --  doesn't handle case if length of coldata doesn't match rest
      of table...should tack nils everywhere to round out the table?"
  (setf (dtable-header dt) (append (dtable-header dt) 
                                (list (read-from-string colname))))
  (setf (dtable-label dt) (append (dtable-label dt) 
     (list colname)))
;;  (setf (dtable-types dt) 
;;     (append (dtable-types dt) (list (type-of (first coldata)))))
  (setf (gethash (read-from-string colname) (dtable-data dt)) coldata) 
  (setf (first (dtable-stats dt)) (length (dtable-header dt)))
  colname
)
;; --------------------------------------------------- add-obs
(defun add-obs (dt obsdata)
"add-obs (dt obsdata) -- add the observation to the end of the dtable dt"
;  (format t "Adding ~A as the ~:R observation~%" obsdata (nobs dt))
  (dotimes (k (length obsdata))
     (vector-push-extend (elt obsdata k) 
        (gethash (elt (dtable-header dt) k) (dtable-data dt))))
  (incf (second (dtable-stats dt))))
;; --------------------------------------------------- del-col
(defun del-col (dt colname)
"del-col (dt colname) -- delete the column colname from dtable dt."
 ; find position of colname, remove that position from header,lable,types.
 ; then remove the sequence in the data hash, and 
 ; set the length to the stats according to length of colname.
;; not yet implemented!!!
 (declare (ignore dt colname))
 nil)
;; --------------------------------------------------- nobs
(defun nobs (dt)
  "nobs (dt) -- return number of observations in table dt. 
  Should return nil if table not found!!"
  (length (gethash (first (dtable-header dt)) (dtable-data dt))))
;; --------------------------------------------------- dtable-export
(defun dtable-export (dt filename)
  "dtable-export (filename)
  writes dtable to filename as csv file "
  (with-open-file (fout filename :direction :output)
    (write-csv-line fout (dtable-label dt))
    (dotimes (k (nobs dt))
       (write-csv-line fout (get-obs dt k)))))
;; --------------------------------------------------- dtabulate
(defun dtabulate (hash &optional label)
  (let ((dt (make-dtable)))
    (setf (dtable-header dt) (keys hash))
    (setf (dtable-label dt) (or label (mapcar #'(lambda (x) (format nil "~A" x))
                                         (keys hash))))
    (setf (dtable-data dt) hash)
   dt))
 ;; --------------------------------------------------- dtable-import
(defun dtable-import (filename)
  "dtable-import (filename)
  reads filename and loads it into a dtable struct which it returns
  "
  (let ((dt (multiple-value-bind (hash label) (csv-to-hash filename)
    (dtabulate hash label))))
  dt))
;; --------------------------------------------------- group
(defun group (table expr)
  "group (table grpcols) -- group the dtable structure by cols,
  save the group hash in the structure and return it.
  Note: grpcols can be functions of the columns. 
  eg (group dt '((> sorate 5)(< 100000 salary)))"
  (dt-hide table nil)
  (let ((grpHash (make-hash-table :test #'equal))
        (dataset (loop for x in expr collect 
                    (gethash x (dtable-data (dt-subset table expr))))))
     (dbg :group "dataset ~A" dataset)
     (dotimes (n (nobs table) grpHash)
         (let ((key (map 'list #'(lambda (x) (elt x n)) dataset)))
            (push n (gethash key grpHash))))
     grpHash))
;; -------------------------------------------------------- get-group
(defun get-group (dt grpcols)
  (gethash (sort-copy grpcols #'string-lessp) (dtable-groups dt)))
;; -------------------------------------------------------- get-group-ids
(defun get-group-ids (dt grpcols)
  (keys (gethash (sort-copy grpcols #'string-lessp) (dtable-groups dt))))
;; -------------------------------------------------------- dt-concat
(defun dt-concat (dt1 dt2)
"(defun dt-concat (dt1 dt2) -- concatenate 2 data sets"
  (let ((header (union (dtable-header dt1) (dtable-header dt2)))
        (data1  (dtable-data dt1)) (data2 (dtable-data dt2))
        (len1   (nobs dt1))    (len2  (nobs dt2)))
    (dtabulate 
      (loadhash (thread header
                 (loop for k in header
          collect (concatenate 'vector (or (gethash k data1) (make-array len1))
                               (or (gethash k data2) (make-array len2)))))))))
;; -------------------------------------------------------- dt-drop
(defun dt-drop (dt cols)
"dt-drop (dt cols) --  drop cols from the dtable dt, return a dtable" 
  (dt-subset dt (set-difference (dtable-header dt) cols)))
;; -------------------------------------------------------- dt-project
(defun dt-project (dt expr &key where by)
"dt-project (dt expr &key by where) -- get the expression from the dtable dt,
return a list of column data." 
  (let ((data (gethash expr (dtable-data (dt-subset dt `(,expr) :where where)))))
    (cond (by (let* ((args (remove-duplicates 
                             (intersection (arguments by) (dtable-header dt))))
                     (grps (group (dt-subset dt args :where where) by))
                     (reshash (loadhash (thread (keys grps) 
                                (loop for grp in (keys grps) collect 
                                  (extract data (gethash grp grps)))))))
                reshash))
          (t data))))
;; -------------------------------------------------------- dt-random-subset
(defun dt-random-subset (dt n)
"dt-random-subset (dt n) -- select n random observations from dt. If n<1, 
then choose that fraction of the total dataset"
  (let ((elems (deal-random n (nobs dt)))
        (data  (dtable-data dt)))
    (dtabulate 
      (loadhash (thread (keys data)
                   (loop for k in (keys data)
                    collect (extract (gethash k data) elems)))))))
;; -------------------------------------------------------- dt-rename
(defun dt-rename (dt new old)
"dt-rename (dt new old) -- still has side-effects!!! needs to copy the data"
  (let* ((dt-cp (dtabulate (dtable-data dt)))
         (hash (dtable-data dt-cp)))
    (mapcar 
      #'(lambda (n o) 
          (when (not (string-equal n o)) 
            (format t "n ~A type of n ~A~%" n (type-of n) )
            (setf (gethash n hash) 
                    (gethash o hash)) (remhash o hash)))
               new old)
    (setf (dtable-header dt-cp) (keys hash))
;    (setf (dtable-label dt) (or label (mapcar #'(lambda (x) (format nil "~A" x))
;                                         (keys hash))))
    dt-cp
))
;; -------------------------------------------------------- dt-join
(defun dt-join (dta dtb match &key matchb 
                              (drop-multiple '(nil nil)) (keep-nomatch '(nil nil)))
 "dt-join (dta dtb match &key matchb 
                           (drop-multiple '(t t)) (keep-nomatch '(nil nil)))
  -- join tables a and b on matching columns. Output a dtable.
  Match is a list of exprs, matchb defaults to match if nil.  "
  ; build a list of output columns based on keep-cols.
  (declare (ignore keep-nomatch))
  (let  ((agroup (group dta match))
         (bgroup (group dtb (or matchb match)))
         (wboc   (append match matchb  ; would-be outcols.
                  (set-difference (append (dtable-header dta) (dtable-header dtb)) 
                                  (append match matchb))))
         (out (make-hash-table :test #'equal))
         (outcols (append match matchb (rename-duplicates 
                   (set-difference (append (dtable-header dta) (dtable-header dtb)) 
                                   (append match matchb)))))
         (adata  (dtable-data dta)) 
         (btable nil) (bdata nil ))
    ; it must be safe to assume that each table had no duplicate column names 
    ; coming in to this routine. Thus there can only be one duplicate possible,
    ; and it must be in table b.
    (setf outcols (mapcar #'(lambda (s) (if (stringp s) (read-from-string s) s))
                             outcols))
    (setf wboc    (mapcar #'(lambda (s) (if (stringp s) (read-from-string s) s))
                             wboc))
    (setf btable (dt-rename dtb outcols wboc))
    (setf bdata  (dtable-data btable))

    (loop for col in outcols do
       (setf (gethash col out) (make-array 0 :adjustable t :fill-pointer t)))
    (loop for k in (keys agroup) do
       (when (gethash k bgroup) 
          (loop for col in outcols do
          (dolist (p-a (if (first drop-multiple) (list (first (gethash k agroup))) 
                                                 (gethash k agroup)))
              (dolist (p-b (if (second drop-multiple) 
                                  (list (first (gethash k bgroup)))
                                  (gethash k bgroup)))
                 (vector-push-extend (progn
                   (if (gethash col adata) (elt (gethash col adata) p-a) 
                                           (elt (gethash col bdata) p-b)))
                  (gethash col out)))))))
    (setf dtb (dt-rename dtb wboc outcols)) ; undoes side effects (i hope)
  (dtabulate out)))
;; -------------------------------------------------------- dt-split
(defun dt-split (dt group label exprs)
"defun dt-split (dt group label exprs) -- a type of transpose
 Need to make sure that the new column names don't contain duplicates. "
  (let* ((dtt (dt-subset dt (append group (list label) exprs)))
         (grp (group dtt group)) (ielem 0)
         (data (dtable-data dtt))
         (labs (remove-duplicates (dt-project dt label) :test #'equal))
         (result (make-hash-table :test #'equal)))
   (dbg :split "group ~A " grp) (dbg :split "keys ~A " (keys grp))
   (loop for g in group do 
       (setf (gethash g result) (make-array (length (keys grp))))) 
   (loop for l in (coerce labs 'list) do
     (loop for e in exprs do
       (let ((c (format nil (format nil "~A_~A" e l))))
         (setf (gethash c result) (make-array (length (keys grp)))))))
   
   (loop for k in (keys grp) do
     (dbg :split "k ~A" k)
     (loop for ig from 0 to (1- (length k)) do
        (setf (elt (gethash (elt group ig) result) ielem) (elt k ig)))
     (loop for v in (gethash k grp) do
        (let ((lab (format nil "~A" (elt (gethash label data) v))))
          (loop for expr in exprs do
             (let ((lab2 (format nil "~A_~A" expr lab)))
                (setf (elt (gethash lab2 result) ielem) 
                      (elt (gethash expr data) v))
             ))))
    (incf ielem))
   (dtabulate result))
)
;; -------------------------------------------------------- dt-subset
(memoize 
  (defun dt-subset (dt expr &key where)
"dt-subset (dt expr &key where) -- get the expression from the dtable dt,
return a dtable of expressions." 
  (dbg :subset "expr ~A where ~A" expr where)
  (let* ((args (remove-duplicates 
                  (intersection (arguments expr) 
                                (dtable-header dt))))
         (cols (progn  ; inefficient, but needed to match obsnums
              (when where (dt-hide dt nil) (dt-hide dt where :if-not t)) 
              (get-cols dt args)))
         (formula nil)
         (fn nil)
         (table 
      (dtabulate 
	 (loadhash (thread expr (loop for x in expr collect
	   (let* ((minargs (intersection (arguments (list x)) (dtable-header dt)))
                  (formula `(lambda ,args 
                                (declare (ignore ,@(set-difference args minargs)))
                                (and ,@minargs ,x))) 
		  (fn (compile nil formula)))
	 (dbg :subset "formula ~A cols ~A" formula cols)
	 (eval (append `(map 'vector ,fn) cols))))) :test #'equal))))
     (dt-hide dt nil)
     table))
  :key #'identity :test #'equal)
;; -------------------------------------------------------- get-col
(defun get-col (dt col &optional (convert t))
  "get-col (dt col &optional (convert t)) -- get the column from the dtable dt,
  return a list of column data. if convert nil, then don't try to convert
  the data to numeric"
  (let ((types (dtable-types dt)) (data (dtable-data dt)))
  (cond ((eq col :n) (gethash col data))
        (convert
           (unless (gethash col types) 
             (setf (gethash col data) (convert-to-num (gethash col data)))
             (setf (gethash col types) 
               (type-of (find-if-not #'null (gethash col data)))))
           (get-col dt col nil))
        ((dtable-hidden dt)
           (extract (gethash col data) (dtable-active dt)))
        (t (gethash col data)))))
;; -------------------------------------------------------- get-cols
(defun get-cols (dt cols)
  "get-cols (dt cols) -- get the columns from the dtable dt,
  return a list of column data. Need to add the by and where keys"
  (cond ((null cols) ()) 
        (t (mapcar 
	   #'(lambda (x) (get-col dt x)) cols))))
;; -------------------------------------------------------- get-col-by
(defun get-col-by (dt col grpcols)
  "get-col-by (dt col grpcols) -- group data by grpcols, 
  return the hash with keys of group ids, values are the values
  of the column col in each group."
  (group dt grpcols)
  (let ((colvals (get-col dt col))(col-by (make-hash-table)))
    (maphash #'(lambda (key val) 
                   (setf (gethash key col-by) (extract colvals val))) 
                   (get-group dt grpcols)) 
    col-by))  
;; -------------------------------------------------------- get-stats-by
(defun get-stats-by (dt col grpcols)
  "get-stats-by (dt col grpcols) -- group data by grpcols, 
  return the hash with keys of group ids, values are the stats of the values
  of the column col in each group."
  (let ((hcols (get-col-by dt col grpcols))
        (hstat (make-hash-table :test #'equal)))
    (maphash #'(lambda (key val) 
                   (setf (gethash key hstat) (simple-stats val))) hcols )
    hstat))  
;; -------------------------------------------------------- get-obs
(defun get-obs (dt n)
 "get the nth observation (all columns) from the dtable as a list
 Should signal an error when past eof!"
 (if (< n (nobs dt))
   (mapcar #'(lambda (x) (elt (get-col dt x) 
                         (elt (dtable-ordering dt) n))) (dtable-header dt))
   nil))
;; -------------------------------------------------------- join
(defun join (dta dtb match-cols keep-cols 
              &optional (drop-multiple '(t t)) (keep-nomatch '(nil nil)))
 "join (dta dtb match-cols keep-cols 
              &optional (drop-multiple '(t t)) (keep-nomatch '(nil nil)))
  -- join tables a and b on matching columns. Output a dtable with columns
  in the keep-cols lists. Match-cols is a list of pairs, keep-cols is a pair
  of lists.
  -- TO DO make function to turn data into a dtable (dtabulate data headers).
           drop the idea of keep-cols -- this should be done by 
                   (dtabulate (dt-project xxx)) (dtabulate (dt-project yyy))
           make match-cols same name by default...eg ( serial (sliderid slidersn) headno)
"
  ; build a list of output columns based on keep-cols.
  (declare (ignore drop-multiple keep-nomatch))
  (let* ((out (make-dtable))
         (amatch (mapcar #'first match-cols))
         (bmatch (mapcar #'second match-cols))
         (agroup (group dta amatch))
         (bgroup (group dtb bmatch))
         (akeep (first keep-cols)) (bkeep (second keep-cols))
         (aobs nil) (bobs nil))
    (mapcar #'(lambda (x) (add-col out x 
        (make-array 0 :adjustable t :fill-pointer 0))) (flatten keep-cols))
  ; loop over keys of agroup
    (maphash #'(lambda (key val) 
      (setf aobs val)
      (setf bobs (gethash key bgroup)) 
    ; here i should mix aobs and bobs in all possible ways...later.
      (add-obs out (flatten (coerce 
        (thread
          (extract (thread-lists (get-cols dta akeep)) aobs)
          (extract (thread-lists (get-cols dtb bkeep)) bobs)) 'list)))
      ; put in the datoura-observation-number column!
      )
      agroup)
  out))
;; -------------------------------------------------------- formula-col
(defun formula-col (dt fn cols)
" formula-col (dt fn cols) -- apply fn to the column list cols in dt,
  return a single vector whose length is (length dt), so to speak."
   (apply #'map 'array fn (get-cols dt cols )))
;; -------------------------------------------------------- bygroups
(defun bygroups (dt)
" bygroups (dt) -- return a list of the names of the groups that exist
  for dtable dt"
  (keys (dtable-groups dt)))
;; -------------------------------------------------------- groupkeys
(defun groupkeys (dt)
" groupkeys (dt) -- print the keys of the groups, and return a list
  of every key inside each group"
  (map 'list 
     #'(lambda (key) 
	  (format t "~A " key)
          (keys (gethash key (dtable-groups dt)))) 
     (keys (dtable-groups dt)))) 
;; -------------------------------------------------------- rehide
(defun rehide (dt seq)
"rehide (dt seq) -- reset hidden obs to previously hidden set"
  (setf (dtable-hidden dt) seq )
  (setf (dtable-active dt) 
     (set-diff-svec (coerce (range (- (nobs dt) 1)) 'vector) seq))
  (values))
;; -------------------------------------------------------- select-all
(defun select-all (dt)
"select-all (dt) -- clear any hidden obs and set all obs active"
  (setf (dtable-hidden dt) () )
  (setf (dtable-active dt) (coerce (range (- (nobs dt) 1)) 'vector))
  (values))
;; -------------------------------------------------------- dt-hide
(defun dt-hide (dt test &key (extend nil) (if-not nil))
" dt-hide (dt test &key extend) -- set the hidden field for the observations
  where the test evaluates true. Use nset-difference with 
  DATOURA-OBSERVATION-NUMBER to get the complement. 
  If keyword arg :extend is true, add cut obs to cut list, otherwise
  replace the cut list."
  (cond ((null test) (select-all dt))
        (t 
  (when if-not (setf extend (not extend)))    ;?? is this right logic?
  (when (and if-not extend) (select-all dt))
  (let ((tocut (coerce (positions #'identity (dt-project dt test)) 'vector))
        (universal (coerce (range (- (nobs dt) 1)) 'vector)))
   (dbg :hide "hide: tocut is ~A" tocut)
   (when if-not (setf tocut (set-diff-svec universal tocut)))
   (if extend 
     (setf (dtable-hidden dt) 
       (delete-duplicates (concatenate 'vector tocut (dtable-hidden dt))))
     ;else
     (setf (dtable-hidden dt) tocut))   
    (setf (dtable-active dt) 
      (set-diff-svec universal (dtable-hidden dt))))
  test
  )))
;; -------------------------------------------------------- dtable-sort
(defun dtable-sort (dt collist &optional template)
  "dtable-sort (dt collist) -- sort the table according to the 
  columns in collist ascending or descending"
  ; parse the collist to determine cols and ascending or descending
  (let ((fnlist nil) (cols nil) (datlist nil) (sortout nil))
    (dolist (elem collist)
      (cond ((consp elem) 
              (setf cols (append cols (list (first elem))))
              (setf fnlist (append fnlist (last elem))))
            ((atom elem) 
              (setf cols (append cols (list elem)))
              (setf fnlist (append fnlist '(<))))
            (t nil)))
  ; extract the data for the columns, plus observation number at end.
;    (setf cols (append cols '(:n)))
    (dbg :sort "cols ~A"  cols)
    (setf datlist (thread-lists (get-cols dt cols)))
  ; call the list-sort-asc-desc function
    (setf sortout (list-sort-asc-desc datlist fnlist template))
  ; set dtable-ordering dt to the observation numbers
    (setf (dtable-ordering dt) 
       (map 'vector #'(lambda (x) (first (last x))) sortout))
  ; return dtable-ordering dt
    ))
;; -------------------------------------------------------- dt-where
(defun dt-where (dt where)
"dt-project (dt where) -- get the observations from the dtable dt,
where the where condition is true. Return a dtable" 
  (dt-subset dt (dtable-header dt) :where where))
;; -------------------------------------------------------- print-dtable
(defun print-dtable (dt s lev)
  "print-dtable (dt s lev) This needs to get alot nicer.
  First, it should use the (dtable-active dt) list to decide which 
  elements to print. Also, isn't there a built-in way to abbreviate 
  the printing of a list?"
  (declare (ignore lev))
  (let ((print-length *print-length*)
        (print-right-margin *print-right-margin*))
    (setf *print-length* *dtable-print-length* 
          *print-right-margin* 500)
    (fresh-line s)
    (format s "#< dtable data structure ~%")
    (format s "Size   : ~D Observations of ~D Variables ~%" 
         (nobs dt) (length (dtable-header dt)))
    (format s "Header : ~A~%" (dtable-label dt))
    (format s "Types  : ~A~%" (dtable-types dt))
    (format s "Data   : ~%" )
    (dolist (elem (dtable-header dt))
      (format s "  ~A -> ~A ~%" elem (get-col dt elem nil)))
    (format s "Properties : ~A  " (dtable-properties dt))
    (format s ">")
    (setf *print-length* print-length 
          *print-right-margin* print-right-margin)))
