;;  graphics.lsp. The Data Touring Machine. Copyright (C) 2004 Steve Eichblatt
;;  Distributed under terms of GNU Public License. Details at end of file.
(unless (find-package :utils) (load (compile-file "utils")))
(defpackage :datoura-graphics
           (:nicknames :graphics)
           (:use :common-lisp :cl-user :datoura-utils)
           (:export #:graphics-object #:copy-graphics-object #:stretch-point
                    #:point #:ps-directive #:line #:stretch-box #:rectangle 
                    #:polygon #:rewind #:ffwd #:show-current #:*graphics-path*
                    #:calc-major-ticks #:legend-entry #:circle
                    #:set-color #:unset-color #:grestore #:make-glist 
                    #:diamond #:glist-push #:glist-nconc
                    #:glist-add #:glist-p #:glist-lst #:glist-append
                    #:text #:axis #:axis-setting #:show))
(in-package :datoura-graphics)

(eval-when (eval compile load)
#+CLISP 
  (setf custom:*warn-on-floating-point-contagion* nil) ; this is clisp-specific.
)

(load "ps-prolog")
(defparameter *postscript-viewer* 
#+cygwin
  (format nil "~A~A " (namestring (user-homedir-pathname)) "csh/my-gv")
#-cygwin "gv -spartan" "The postscript viewer program")

(defparameter *gv-pid* nil "The process id of the postscript viewer")

(defun set-graphics-path (path)
  (defparameter *graphics-path* path
     "Directory where the eps files generated are stored")
  (defparameter *graphics-terminal*  
     (merge-pathnames *graphics-path* "terminal.eps")
     "The file to write graphics output"))

(set-graphics-path 
#-(or gcl win32)
  (ensure-directories-exist 
    (pathname (concatenate 'string (namestring (user-homedir-pathname))
    "/graph/")))
#+(or gcl win32) (user-homedir-pathname)
)

(defparameter *graphics-options* 
  `((xsize . ,(* .8 480)) (ysize . ,(* .8 296))
   (axes . (:x0 :y0)) (show-axes . t) (margin . ((:rx . (0 1)) (:ry . (0 1))))))
(defparameter *text-options* '((text-angle . 0) (text-justification . left)
   (text-size . 15)))
(defparameter *point-options* '((point-symbol . :mdpt)))
(defparameter *ps-options* '(()))
(defparameter *line-options* '((coords . axis)))
(defparameter *circle-options* '(()))
(defparameter *polygon-options* '((fill . nil)))
(defparameter *diamond-options* '((diamond-command . "diamond")))
(defparameter *rectangle-options* '((rec-type . ()) (coords . axis)))
(defparameter *default-margin* '(0.15 0.85))
(defparameter *glist-print-length* 10)
(defconstant *tiny-number* (* 1000 least-positive-short-float) 
"Extent of a point in datoura space. I don't know why it has to be so big.")
(defconstant *default-major-ticks* 8
 "How many tick marks to try to put on an axis")
(defconstant *major-tick-len* 7)
(defconstant *minor-tick-len* 5)
(defparameter *axis-options* '((label-precision . 4)))
(defparameter *axis-marker-font* "swiss-small")
(defparameter *x-marker-justification* :center-top)
(defparameter *y-marker-justification* :right-center)
(defparameter *legend-options* '((leg-xpos . .18)(leg-ypos . .9)
                                 (leg-dx   .  0)(leg-dy   . -.05)
                                 (leg-font . "8 fixed-n")))
(defparameter *plot-counter* 0) (defparameter *current-plot* 0)
;; ------------------------------------------------------ gr-obj-print
(defun gr-obj-print (gobj s lev)
"gr-ojb-print (gobj s lev) "
  (declare (ignore lev))
  (let ((print-length *print-length*)
        (print-right-margin *print-right-margin*))
    (setf *print-length* 10
          *print-right-margin* 500)
    (fresh-line s)
    (format s "~&#< graphics-object ~A ~A ~A ~%~A" 
       (gr-obj-type gobj) (gr-obj-value gobj) (gr-obj-options gobj)
       (gr-obj-location gobj))
    (format s ">")
    (setf *print-length* print-length 
          *print-right-margin* print-right-margin)))
;; ------------------------------------------------------ graphics-object
(defstruct (graphics-object (:conc-name gr-obj-) 
              (:print-function gr-obj-print) (:copier copy-graphics-object))
  "graphics-object -- The structure to contain graphics primitives, such as 
   points, lines, text etc"
   type 
   location
   coords
   value
   options 
   bbox
   display-fn 
)
;; ------------------------------------------------------ copy-graphics-object
(defun copy-graphics-object (obj)
"copy-graphics-object (obj) -- returns a 'deep copy' of the graphics
object obj. This allows manipulation of co-ordinates of the copy 
without affecting the original. Otherwise, a glist can only be scaled once."
 (declare (optimize (safety 3)))
 (let ((cp (make-graphics-object)))
   (setf (gr-obj-type cp) (gr-obj-type obj))
   (setf (gr-obj-location cp) (copy-tree (coerce (gr-obj-location obj) 'list)))
   (setf (gr-obj-coords cp) (gr-obj-coords obj))
   (setf (gr-obj-value cp) (gr-obj-value obj))
   (setf (gr-obj-options cp) (copy-alist (gr-obj-options obj)))
   (setf (gr-obj-bbox cp) (copy-tree (gr-obj-bbox obj)))
   (setf (gr-obj-display-fn cp) (gr-obj-display-fn obj))
   cp))
;; ------------------------------------------------------ glist
(defstruct (glist (:print-function glist-print)
                  (:copier copy-glist))
   lst   ; the list of graphics objects
)
;; ------------------------------------------------------ copy-glist
(defun copy-glist (obj)
"copy-glist (obj) -- returns a 'deep copy' of the glist
This allows manipulation of co-ordinates of the copy 
without affecting the original. Otherwise, a glist can only be scaled once."
 (declare (optimize (safety 3)))
 (make-glist :lst 
   (mapcar #'copy-graphics-object (glist-lst (glist-append obj)))))
;; ------------------------------------------------------ glist-print
(defun glist-print (gl s lev)
"glist-print (gl s lev) "
  (declare (ignore lev))
  (let ((print-length *print-length*)
        (print-right-margin *print-right-margin*))
    (setf *print-length* *glist-print-length*
          *print-right-margin* nil)
    (fresh-line s)
    (format s "~&#< glist data structure ")
    (format s ": ~D entries " (length (glist-lst gl)))
    (dbg :glist-print "glist ~A" (glist-lst gl))
    (format s "~&~A " 
       (category-freq 
         (mapcar #'(lambda (x) (gr-obj-type x)) (glist-lst gl))))
    (dolist (elem (glist-lst gl))
      (when (member (symbol-name (gr-obj-type elem)) 
         (mapcar #'symbol-name *dbg-ids*) :test #'string-equal)
        (format s "~&~A" elem)))
    (format s ">~%")
    (setf *print-length* print-length 
          *print-right-margin* print-right-margin)))
;; ------------------------------------------------------ glist-nconc
(defun glist-nconc (&rest objs)
"glist-nconc (&rest objs) -- returns glist object of objs, whether
they are glists, lists of or atomic gr-objs. This is a faster, destructive
version of glist-append"
 (make-glist :lst 
    (mapcan 
      #'(lambda (x) 
          (cond ((glist-p  x) (glist-lst x))
                ((graphics-object-p x) (list x))
                ((consp x) (remove-if-not #'graphics-object-p x))
                (t ()))) 
         (flatten objs))))
;; ------------------------------------------------------ glist-append
(defun glist-append (&rest objs)
"glist-append (&rest objs) -- returns glist object of objs, whether
they are glists, lists of or atomic gr-objs"
 (declare (optimize (safety 3)))
 (make-glist :lst 
   (let ((result ())
         (gl (mapcar 
           #'(lambda (x) 
               (cond ((glist-p  x) (glist-lst x))
                     ((graphics-object-p x) (list x))
                     ((consp x) (remove-if-not #'graphics-object-p x))
                     (t ()))) 
              (flatten objs))))
      (dolist (elem gl result) (setf result (append result elem))))))
;; ------------------------------------------------------ glist-push
(defmacro glist-push (gl &rest new-objs)
"glist-push (gl &rest new-objs) -- push graphics objects on 
THE BEGINNING of gl"
  `(setf ,gl (glist-nconc ,@new-objs ,gl)))
;; ------------------------------------------------------ glist-add
(defmacro glist-add (gl &rest new-objs)
"glist-add (gl &rest new-objs) -- adds graphics objects TO THE END of gl"
  `(setf ,gl (glist-nconc ,gl ,@new-objs)))
;; ------------------------------------------------------ axis-object
(defstruct (axis-object (:conc-name ax-obj-))
  "axis-object -- The structure to contain description of an axis"
   name
   label
   min                      
   max                    
   transform-fn      ; convert from axis -> absolute coords
   angle             ; x -> 0, y -> 90.
  (location .15)     ; relative posn of axis in the other dimension
   margin            ; (min max) dist. in relative coords from edges 
   major-ticks       ; distance between labeled ticks in axis coords
  (major-offset 0)   ; if major-tick=10 & major-offset=1, ticks at 1,11,..
  (n-minor-ticks 1)  ; number of minor ticks between major ticks
   options                   
   glist             ; gr-obj's to compose axis.
   display-fn        ; function which draws axis in ps.
  (exp 0)            ; to keep the labels on the axis to reasonable size.
  (marker-label-fn   ; the function to convert locations to labels.
    #'(lambda (xval &optional (exp 0)) 
        (format nil "~A" (float (round-to-nearest (/ xval (expt 10 exp))
            (/ 1 (expt 10 (assoc-val 'label-precision *axis-options*))))))))
)
;; ------------------------------------------------------ axis-exp
(defun axis-exp (min max)
"axis-exp (min max) -- Calculate the power of 10 to scale the 
axis labels by in order to keep them between .001 and 10000"
  (let ((div (expt 10 (assoc-val 'label-precision *axis-options*)))
        (mul (expt 10 (- (assoc-val 'label-precision *axis-options*))))
        (amx (abs max))
        (amn (abs min)))
  (cond ((or (> amx div) (> amn div)) 
           (floor-to-nearest (log (max amn amx) 10) 3))
        ((and (< amx mul) (< amn mul)) 
           (ceiling-to-nearest (log (max amn amx) 10) 3))
         (t 0))))
;; ---------------------------------------------------- calc-major-ticks
(defun calc-major-ticks (min max &key (nticks *default-major-ticks*))
"defun calc-major-ticks (min max &key (nticks *default-major-ticks*))
-- Calculate the major tick spacing for the axis with range min to max 
   trying to get nticks tick marks. The spacing will be a round number
   beginning with 1,2 or 5"
  (if (= max min) .001
    (let* ((range (abs (- max min))) ; paranoid.
           (exponent (floor (log range 10)))
           (normRange (float (/ range (expt 10 exponent)))))
       (cond ((plusp (round-to-nearest normRange nticks)) 
                 (expt 10 exponent))    
             ((plusp (round-to-nearest (* 2 normRange) nticks))
                 (* 0.5 (expt 10 exponent)))
             ((plusp (round-to-nearest (* 5 normRange) nticks))
                 (* 0.2 (expt 10 exponent)))
             (t 1)))
))
;; ------------------------------------------------------ axis-tick-lines
(defun axis-tick-lines (begin location angle tick-len)
  (case angle
  ((0)  (thread (thread begin (make-list (length begin) 
                              :initial-element location))
                (thread begin (make-list (length begin) 
                              :initial-element (- location tick-len)))))
  ((90) (thread (thread (make-list (length begin) 
                        :initial-element location) begin)
                (thread (make-list (length begin) 
                        :initial-element (- location tick-len)) begin )))))
;; ----------------------------------------------------- axis-tick-locations
(defun axis-tick-locations (obj)
"axis-tick-locations (obj) -- return a list of points where 
to start the ticks from, suitable for showing with line."
  (let* ((min (ax-obj-min obj)) (max (ax-obj-max obj))
        (major (or (ax-obj-major-ticks obj) (calc-major-ticks min max)))
        (minor (1+ (ax-obj-n-minor-ticks obj)))
        (minor-space (/ major minor))
        (major-locations (range (ceiling-to-nearest min major) 
                                 max major)) 
        (minor-locations (range (ceiling-to-nearest min minor-space) 
                                 max minor-space)))
  (values major-locations minor-locations)
  )) 
;; ------------------------------------------------------ make-axis-glist
(defun make-axis-glist (axobj &optional (options ()))
"make-axis-glist (axobj &optional (options ())) -- Make a glist, suitable 
for showing with show, that will draw the axis-object axobj"
 (when (assoc-val 'show-axes options)
  (let* ((goptions (alist-join options *graphics-options*))
        (name (ax-obj-name axobj)) 
        (angle (ax-obj-angle axobj)) 
        (xsize (assoc-val 'xsize goptions)) (ysize (assoc-val 'ysize goptions))
        (min (ax-obj-min axobj)) (max (ax-obj-max axobj)) (range 0)
        (location (ax-obj-location axobj)) 
        (second-axis (> location .5))
        (major-tick-len (if second-axis (- *major-tick-len*) *major-tick-len*))
        (minor-tick-len (if second-axis (- *minor-tick-len*) *minor-tick-len*))
        (yjust (or (assoc-val 'y-marker-justification goptions) 
                   (if second-axis :left-center :right-center)))
        (xjust (or (assoc-val 'x-marker-justification goptions) 
                   (if second-axis :center :center-top)))
        (gl (make-glist)))
  (case name
    ((:rx :ry :ax :ay) ())     ;; don't show relative or absolute axes.
    (otherwise   
     (setf range (- max min))
     (dbg :axisT "major-ticks is ~A" (ax-obj-major-ticks axobj))
     (setf (ax-obj-exp axobj) (axis-exp min max))
     (dbg :axis "Axis name ~A exponent is ~A" name (ax-obj-exp axobj))
     (dbg :axis "Axis name ~A min ~A max ~A " name min max )
     (dbg :aaa "xsize ~A ysize ~A" xsize ysize)
     (multiple-value-bind (major-loc minor-loc) (axis-tick-locations axobj)
      (case angle
         ((0) (setf major-tick-len (/ major-tick-len ysize)) 
              (setf minor-tick-len (/ minor-tick-len ysize)) 
              (setf (ax-obj-major-ticks axobj) 
                (or (assoc-val 'x-major-ticks goptions) 
                    (ax-obj-major-ticks axobj) (calc-major-ticks min max)))
              (setf (ax-obj-n-minor-ticks axobj) 
                (or (assoc-val 'x-minor-ticks goptions) 
                    (ax-obj-n-minor-ticks axobj)))
              (dbg :aaa "major-tick-len ~A" major-tick-len)
              (glist-add gl (mapcar #'(lambda (x) 
                   (line x `(axes . (,name :ry))))    ;; draw major ticks 
                      (axis-tick-lines major-loc location angle 
                       major-tick-len)))
              (glist-add gl (mapcar #'(lambda (x)  ;; axis markers
                 (text x (- location (* 1.2 major-tick-len))
                   (funcall (ax-obj-marker-label-fn axobj) x (ax-obj-exp axobj)) 
                    `((text-justification . ,xjust) (coords . axis) 
                      (axes . (,name :ry)) (text-font . ,*axis-marker-font*)))) 
                       major-loc))
              (glist-add gl (mapcar #'(lambda (x) 
                   (line x `(axes . (,name :ry))))    ;; draw minor ticks
                      (axis-tick-lines minor-loc location angle 
                       minor-tick-len)))
              (glist-push gl (text (/ (+ max min) 2)         ;; axis label
                           (- location (* 4 major-tick-len))
                           (ax-obj-label axobj) `((text-justification . center)
                           (coords . axis) (axes . (,name :ry)))))
              (unless (= (ax-obj-exp axobj) 0)
                 (glist-push gl (text max 0   ;; axis exponent
                      (with-output-to-string (xx) (format xx "*10^~D" 
                           (ax-obj-exp axobj))) `((text-justification . right)
                           (text-font . "swiss")
                           (coords . axis) (axes . (,name :ry))))))
              (glist-push gl 
                (line (list (list min location) (list max location))
                         `(axes . (,name :ry))))  ;; draw axis line
         ) 
         ((90) (setf major-tick-len (/ major-tick-len xsize))
               (setf minor-tick-len (/ minor-tick-len xsize)) 
               (dbg :aaa "major-tick-len ~A" major-tick-len)
               (setf (ax-obj-major-ticks axobj) 
                 (or (assoc-val 'y-major-ticks goptions) 
                     (ax-obj-major-ticks axobj) (calc-major-ticks min max)))
               (setf (ax-obj-n-minor-ticks axobj) 
                 (or (assoc-val 'y-minor-ticks goptions) 
                     (ax-obj-n-minor-ticks axobj)))
                (glist-add gl (mapcar #'(lambda (x)     ;; draw major ticks
                   (line x `(axes . (:rx ,name)))) 
                      (axis-tick-lines major-loc location angle 
                       major-tick-len)))
               (glist-add gl (mapcar #'(lambda (x)  ;; axis markers
                 (text (- location (* 1.2 major-tick-len)) x 
                    (funcall (ax-obj-marker-label-fn axobj) x (ax-obj-exp axobj)) 
                    `((text-justification . ,yjust) (coords . axis) 
                      (axes . (:rx ,name)) (text-font . ,*axis-marker-font*)))) 
                     major-loc))
               (glist-add gl (mapcar #'(lambda (x) 
                   (line x `(axes . (:rx ,name))))  ;; draw minor ticks
                      (axis-tick-lines minor-loc location angle 
                        (* .7 minor-tick-len))))
               (glist-push gl (text (- location (* 6 major-tick-len))
                           (/ (+ max min) 2)       ;; axis label
                           (ax-obj-label axobj) 
                           `((text-justification . center)
                           (coords . axis) (text-angle . 90) 
                           (axes . (:rx ,name)))))
               (unless (= (ax-obj-exp axobj) 0)    ;; axis exponent
                 (glist-push gl (text (- location (* 6 major-tick-len))
                                      (* max .97)  
                      (with-output-to-string (xx) (format xx "*10^~D" 
                           (ax-obj-exp axobj))) 
                      `((text-justification . right-center)
                       (text-font . "swiss-large") (text-angle . 90)
                       (coords . axis) (axes . (:rx ,name))))))
                (glist-push gl 
                  (line (list (list location min) (list location max))
                         `(axes . (:rx ,name))))  ;; draw axis line
         )
         (otherwise ()))))
  ) 
  (dbg :axis "in make-axis-glist gl is ~A" gl)
  gl)))
;; ------------------------------------------------------ set-color
(defun set-color (color)
  (ps-directive (format nil "~{~6,4F ~} setrgbcolor" color)))
;; ------------------------------------------------------ unset-color
(defun unset-color ()
  (set-color '(0 0 0)))
;; ------------------------------------------------------ grestore
(defun grestore ()
  (list (ps-directive "grestore")))
;; ------------------------------------------------------ axis-setting
(defun axis-setting (name &key label location major-ticks 
                               n-minor-ticks marker-label-fn options )
  "axis-setting (name &key label location 
                           major-ticks n-minor-ticks options ) --
  create a graphics-object that will be used to modify the axis with
  name 'name' after graphics-compile"
  (make-graphics-object :type 'axis-setting 
   :display-fn #'(lambda (axes) 
   (let ((ax (assoc-val name axes)))
     (dbg :axis "~&axis-setting: ~A" ax)
     (case ax 
       ((nil) ())
       (otherwise
        (when label (setf (ax-obj-label ax) label))
        (when location (setf (ax-obj-location ax) location))
        (when major-ticks (setf (ax-obj-major-ticks ax) major-ticks))
        (when n-minor-ticks (setf (ax-obj-n-minor-ticks ax) n-minor-ticks))
        (when marker-label-fn (setf (ax-obj-marker-label-fn ax)
                                     marker-label-fn))
        (when options (setf (ax-obj-options ax) options))))))))
;; ------------------------------------------------------ make-legend-glist
(defun make-legend-glist (gl &optional (options ()))
"make-legend-glist (lst &optional (options ())) -- make a glist to show
the legend entries "
  (let ((gl-out (make-glist))
        (xpos (assoc-val 'leg-xpos (alist-join options *legend-options*)))
        (ypos (assoc-val 'leg-ypos (alist-join options *legend-options*)))
        (dx   (assoc-val 'leg-dx   (alist-join options *legend-options*)))
        (dy   (assoc-val 'leg-dy   (alist-join options *legend-options*)))
        (font (assoc-val 'leg-font (alist-join options *legend-options*)))
        (lst (remove-if-not 
           #'(lambda (x) 
               (equal 'legend-entry (gr-obj-type x))) (glist-lst gl))))
    (dbg :legend "make-legend-glist ~A" lst)
    (dolist (entry lst)
      (glist-add gl-out (set-color (or (assoc-val 'rgbcolor 
                                        (gr-obj-options entry)) '(0 0 0))))
      (glist-add gl-out 
          (point (- xpos .008) ypos 
             `((point-symbol . ,(assoc-val 'symbol (gr-obj-options entry)))
               (coords . relative))))
      (glist-add gl-out 
          (text xpos ypos (gr-obj-value entry) 
             `((text-justification . left-center) (text-font . ,font)
             (coords . relative))))
      (glist-add gl-out (unset-color))
      (incf xpos dx) (incf ypos dy))
    gl-out))
;; ------------------------------------------------------ legend-entry
(defun legend-entry (label &key (symbol :smsqe) (rgbcolor '(0 0 0)))
"legend-entry (name &key symbol rgbcolor label) -- Create a graphics object
as an element to the plot legend"
  (make-graphics-object :type 'legend-entry
   :display-fn ()  ; not shown individually, but by (make-legend-glist)
   :value label
   :options `((symbol . ,symbol)(rgbcolor . ,rgbcolor))
   ))
;; ------------------------------------------------------ stretch-point
(defun stretch-point (x y &optional (options ()))
  "stretch-point (x y &optional (options ())) --
  return a graphics object invisible point which will reserve space
  on the axes, but will not be shown"
  (if (not (and x y)) 
    nil
    (let ((goptions (alist-join options *point-options*)))
      (make-graphics-object :type 'point :location (list x y) 
       :coords 'axis
       :options goptions :bbox `((,x ,y) 
             (,(+ *tiny-number* x) ,(+ *tiny-number*  y)))
       :display-fn #'(lambda (strm gelem) (declare (ignore strm gelem)) ()) 
       ))))
;; ------------------------------------------------------ point
(defun point (x y &optional (options ()))
  "point (x y &optional (options ())) --
  return a graphics object point suitable for use in show."
  (if (not (and x y)) 
    nil
    (let ((goptions (alist-join options *point-options*)))
      (make-graphics-object :type 'point :location (list x y) 
       :coords (or (assoc-val 'coords goptions) 'axis)
       :options goptions :bbox `((,x ,y) 
                (,(+ *tiny-number* x) 
                 ,(+ *tiny-number* y)))
       :display-fn #'show-point ))))
;; ------------------------------------------------------ show-point
(defun show-point (strm gelem)
  "show-point (strm gelem)
  write a ps description of a point to str"
  (let* ((options  (gr-obj-options gelem))
         (x (first (gr-obj-location gelem)))
         (y (second (gr-obj-location gelem)))
         (point-cmd (string-downcase 
                    (symbol-name (assoc-val 'point-symbol options))))) 
    (fresh-line strm)
    (format strm "~10,5F ~10,5F ~A " x y point-cmd))) 
;; ------------------------------------------------------ ps-directive
(defun ps-directive (value &optional options)
  "ps-directive (value &optional options)
  create a graphics object that passes a postscript string directly 
  to the output file."
  (let ((goptions (alist-join options *ps-options*)))
     (make-graphics-object :type 'ps-directive :value value
     :options goptions
     :display-fn #'show-ps-directive)))
;; ------------------------------------------------------ show-ps-directive
(defun show-ps-directive (strm gelem)
  "show-ps-directive (strm gelem)
  pass the postscript directive to the output stream"
    (fresh-line strm)
;    (format strm (gr-obj-value gelem)))  ; this allows lisp format directives. 
    (format strm " ~A " (gr-obj-value gelem)))  ; this is safer
;; ------------------------------------------------------ line
(defun line (vtx-list &optional (options ()))
  "line (vtx-list &optional options) --
  return a graphics object polyline suitable for use in show."
  (let* ((goptions (alist-join options *line-options*))
        (xs (map 'list #'elt-0 vtx-list))
        (ys (map 'list #'elt-1 vtx-list))
        (sortxs (sort xs #'<))
        (sortys (sort ys #'<)))
    (make-graphics-object :type 'line :location vtx-list
     :coords (or (assoc-val 'coords goptions) 'axis)
     :options goptions 
     :bbox `((,(elt sortxs 0) ,(elt sortys 0))
             (,(elt sortxs (1- (length xs))) 
              ,(elt sortys (1- (length ys)))))
     :display-fn #'show-line )))
;; ------------------------------------------------------ show-line
(defun show-line (strm gelem)
  "show-line (strm x y &optional (options ()))
  write a ps description of a polyline to str"
  (let* ((options  (gr-obj-options gelem))
         (vtx-list (gr-obj-location gelem)))
    (declare (ignore options))  ;; for now.
    (fresh-line strm)
    (format strm "~{ ~10,5F ~}moveto~%" (elt vtx-list 0)) 
    (do ((k 1 (+ k 1))) 
        ((= k (length vtx-list)) ())
      (format strm "~{ ~10,5F ~}lineto~%" (elt vtx-list k)))
    (format strm "stroke~%")))
;; ------------------------------------------------------ circle
(defun circle (x y r &optional (options ()))
"circle (x y r &optional (options ())) --
  return a graphics object circle suitable for use in show."
  (let ((goptions (alist-join options *circle-options*))
        (corners `((,(- x r) ,(- y r)) (,(+ x r) ,(+ y r)))))
    (make-graphics-object :type 'circle
     :location `(,x ,y ,r)
     :coords (or (assoc-val 'coords goptions) 'axis)
     :options goptions 
     :bbox corners
     :display-fn #'show-circle )))
;; ------------------------------------------------------ show-circle
(defun show-circle (strm gelem)
  "show-circle (strm gelem)
  write a ps description of a circle to strm"
  (let* ((options  (gr-obj-options gelem))
         (location (gr-obj-location gelem)) 
         (circle-cmd (if (assoc-val 'fill options) "disk" "circle")))
    (fresh-line strm)
    (format strm "~{ ~10,5F ~}~A~%" location circle-cmd)))
;; ------------------------------------------------------ polygon
(defun polygon (vtx-list &optional (options ()))
  "polygon (vtx-list &optional options) -- 
    valid option: fill default non fill.
  return a graphics object polygon suitable for use in show."
  (let* ((goptions (alist-join options *polygon-options*))
        (xs (map 'list #'(lambda(x) (elt x 0)) vtx-list))
        (ys (map 'list #'(lambda(x) (elt x 1)) vtx-list))
        (sortxs (sort xs #'<))
        (sortys (sort ys #'<)))
    (make-graphics-object :type 'polygon :location vtx-list
     :coords (or (assoc-val 'coords goptions) 'axis)
     :options goptions 
     :bbox `((,(elt sortxs 0) ,(elt sortys 0))
             (,(elt sortxs (1- (length xs))) 
              ,(elt sortys (1- (length ys)))))
     :display-fn #'show-polygon)))
;; ------------------------------------------------------ show-polygon
(defun show-polygon (strm gelem)
  "show-polygon (strm gelem)
  write a ps description of a polygon to str"
  (let* ((options  (gr-obj-options gelem))
         (vtx-list (gr-obj-location gelem)))
    (fresh-line strm)
    (format strm "newpath ~{ ~10,5F ~}moveto~%" (elt vtx-list 0)) 
    (do ((k 1 (+ k 1))) 
        ((= k (length vtx-list)) ())
      (format strm "~{ ~10,5F ~}lineto~%" (elt vtx-list k)))
    (format strm "closepath ~A ~%" 
       (if (assoc-val 'fill options) "fill" "stroke"))))
;; ------------------------------------------------------ rectangle
(defun rectangle (xywh &optional (options ()))
  "rectangle (xywh &optional options) --
  return a graphics object rectangle suitable for use in show.
  corners are of form ((llx lly) (width height)) which is 
  natural for postscript."
  (let ((goptions (alist-join options *rectangle-options*))
        (corners (list (first xywh) 
                 (mapcar #'+ (first xywh) (second xywh)))))
    (make-graphics-object :type 'rectangle 
     :location corners    ; corners are what i need for xforming.
     :coords (or (assoc-val 'coords goptions) 'axis)
     :options goptions 
     :bbox corners
     :display-fn #'show-rectangle )))
;; ------------------------------------------------------ show-rectangle
(defun show-rectangle (strm gelem)
  "show-rectangle (strm x y &optional (options ()))
  write a ps description of a rectangle to str"
  (let* ((options  (gr-obj-options gelem))
         (location (gr-obj-location gelem)) 
         (pspos (list (first location) 
                      (mapcar #'- (second location) (first location))))
         (rec-cmd (or (assoc-val 'rec-type options) "rectstroke")))
    (fresh-line strm)
    (format strm "~{ ~10,5F ~}~A~%" (flatten pspos) 
         rec-cmd)))
;; ------------------------------------------------------ text
(defun text (x y text &optional (options ()))
  "text (x y text &optional options) --
  return a graphics object text suitable for use in show."
  (let ((goptions (alist-join options *text-options*)))
    (make-graphics-object :type 'text :location (list x y) 
     :coords (or (assoc-val 'coords goptions) 'absolute)
     :value text
     :options goptions :bbox `((,x ,y) 
                (,(+ *tiny-number* x) 
                 ,(+ *tiny-number* y)))
     :display-fn #'show-text )))
;; ------------------------------------------------------ show-text
(defun show-text (strm gelem)
  (let* ((options (gr-obj-options gelem))
         (x (first (gr-obj-location gelem)))
         (y (second (gr-obj-location gelem)))
         (angle (assoc-val 'text-angle options))
         (font (assoc-val 'text-font options))
         (text (gr-obj-value gelem))
         (text-cmd (assoc-val (assoc-val 'text-justification options) 
                 '((left . "tlj") (right . "trj") 
                   (center . "tcj") (left-center . "tlc")
                   (right-center . "trc") (center-center . "tcc")
                   (left-top . "tlt") (right-top . "trt") 
                   (center-top . "tct"))))) 
    (cond ((null text) ())
          (t 
    (fresh-line strm)
    (when font (format strm "gsave ~A " font))
    (format strm "~10,5F ~10,5F ~10,5F (~A) ~A " x y angle text text-cmd)
    (when font (format strm " grestore ~%")))))) 
;; ------------------------------------------------------ diamond
(defun diamond (xywh &optional (options ()))
  "diamond (xywh &optional (options ())) --
  return a graphics object diamond suitable for use in show."
  (let* ((goptions (alist-join options *diamond-options*))
        (x  (caar xywh))        (y  (cadar xywh))
        (dx (/ (caadr xywh) 2)) (dy (/ (cadadr xywh) 2))
        (corners `((,(- x dx) ,(- y dy))
                   (,(+ x dx) ,(+ y dy))))) ; for transforming
    (make-graphics-object :type 'diamond 
     :location corners
     :coords (or (assoc-val 'coords goptions) 'axis)
     :options goptions 
     :bbox corners
     :display-fn #'show-diamond )))
;; ------------------------------------------------------ show-diamond
(defun show-diamond (strm gelem)
  "show-diamond (strm gelem)
  write a ps description of a diamond to strm"
  (let* ((options  (gr-obj-options gelem))
         (loc      (gr-obj-location gelem))
         (x  (/ (+ (caar loc)  (caadr  loc)) 2))
         (y  (/ (+ (cadar loc) (cadadr loc)) 2))
         (dx    (- (caadr loc) (caar  loc)))
         (dy    (- (cadadr loc)(cadar loc))))
    (declare (ignore options))
    (fresh-line strm)
    (dbg :diamond "~10,5F ~10,5F ~10,5F ~10,5F diamond" x y dx dy )
    (format strm "~10,5F ~10,5F ~10,5F ~10,5F diamond" x y dx dy ))) 
;; ------------------------------------------------------ graphics-compile
(defun graphics-compile (gl &optional (options ()))
  "graphics-compile (gl &optional (options ())) -- 
  determine the min and max of the x and y axis for axis co-ordinate
  systems. Return (xfactor xoffset yfactor yoffset)"
  ;; note -- this needs to be upgraded to handle: 
  ;; Log plots and 
  ;; multiple axes (eg. left and right, or top and bottom)
  ;; should return something like: 
  ;; ((bottom . #'xtransform) (left . #'ytransform))
  (let* ((axes ()) 
         (bbhx (make-hash-table :test #'equal))
         (bbhy (make-hash-table :test #'equal))
         (xsize (assoc-val 'xsize options))   
         (ysize (assoc-val 'ysize options))   
         (glst  (glist-lst gl))
         (scalable-objects 
           (remove-if-not #'(lambda (g) 
               (or (name-of (gr-obj-coords g) 'axis) 
                   (name-of (gr-obj-coords g) 'relative))) glst)))
    ;; make the relative axes rx and ry in all cases.
    (push '(0 1) (gethash :rx bbhx))
    (push '(0 1) (gethash :ry bbhy))
    (push `(0 ,xsize) (gethash :ax bbhx))
    (push `(0 ,ysize) (gethash :ay bbhy))
    (dolist (obj scalable-objects)
       (when (name-of (gr-obj-coords obj) 'axis)
         (let ((axis 
                (assoc-val 'axes (alist-join (gr-obj-options obj) 
                                           *graphics-options*))))
                (push (mapcar #'first (gr-obj-bbox obj)) 
                      (gethash (first axis) bbhx))
                (push (mapcar #'second (gr-obj-bbox obj)) 
                      (gethash (second axis) bbhy)))))
    (dbg :axis "~%bbhx ~A ~%bbhy ~A" bbhx bbhy)
    (let ((margin (assoc-val 'margin options)))
    ; note -- scale-axes does a linear transformation...
    ; a new function needed if axes are logarithmic. 
       (dbg :compile "margin ~A~%" margin)
       (setf axes (append (scale-axes bbhx xsize margin 0) axes))
       (setf axes (append (scale-axes bbhy ysize margin 90) axes))
    ; apply modifications to axes from glist's axis-settings
    (dolist (obj (remove-if-not #'(lambda (g) 
                         (name-of (gr-obj-type g) 'axis-setting)) glst))
       (funcall (gr-obj-display-fn obj) axes))
    (dbg :compile "axes ~A " axes)
    ) axes)) 
;; ------------------------------------------------------ scale-axes
(defun scale-axes (bbox-hash xsz margin angle)
"scale-axes (bbox-hash xsize margin angle) -- return an alist of 
closures that will linearly scale axes according to the window size and 
bounds"
 (let ((axes ()))
   (maphash #'(lambda (key val)
     (dbg :axis "margin ~A~%" margin)
     (dbg :axis "key ~A~%" key)
     (let* ((mrg (or (assoc-val key margin) *default-margin*))
            (x0 (if (or (eq key :ax) (eq key :ay)
                        (eq key :rx) (eq key :ry)) 
                 0 
                 (* xsz (first mrg))))
            (x1 (if (or (eq key :ax) (eq key :ay)
                        (eq key :rx) (eq key :ry)) 
                 xsz 
                 (* xsz (second mrg))))
           (fval (sort (flatten val) #'<))
           (xmin (elt fval 0))
           (xmax (elt fval (1- (length fval))))
           (xfactor 0) (xoffset 0)
           (xsize (- x1 x0)))
       (dbg :axis "~A size ~10,F min ~10,5F max ~10,5F x0 ~10,5F ~%" 
           key xsize xmax xmin x0)
       (setf xfactor (/ xsize (- xmax xmin)))
       (setf xoffset (+ x0 (- (* xfactor xmin))))
       (dbg :axis "~A size ~10,F xfactor ~10,5F xoffset ~10,5F~%" 
           key xsize xfactor xoffset)
       (setf axes (acons key 
         (make-axis-object :name key :min xmin :max xmax 
         :angle angle :margin mrg
         :transform-fn #'(lambda (x) (+ xoffset (* x xfactor)))) 
             axes)))) bbox-hash)
 axes))
;; ------------------------------------------------------ graphics-transform
(defun graphics-transform (gl 
                  &optional (gaxes (()()()())) (options ()))
  "graphics-transform (gl &optional (options ())) -- 
  set the locations of all objects in gl to absolute co-ordinates."
 (let ((goptions (alist-join options *graphics-options*)))
 ;; note -- if location is a list of co-ordinates, as for a line, then 
 ;; map over it. if not, just apply the funtion.
   (dolist (gelem (glist-lst gl)) 
     (let* ((location (gr-obj-location gelem))
            (coords (gr-obj-coords gelem))
            (axes (if (name-of coords 'relative) '(:rx ry) 
              (assoc-val 'axes 
                 (alist-join (gr-obj-options gelem) goptions)))))
       (dbg :xform "coords is ~A~%" coords)
       (when (or (name-of coords 'relative) (name-of coords 'axis))
            (if (atom (elt location 0))
              (dotimes (k 2) (setf (elt location k) (funcall 
                    (ax-obj-transform-fn (assoc-val (elt axes k) gaxes)) 
                         (elt location k)))
                    (dbg :xform "axis is ~A" (elt axes k)))
              ; else
              (dotimes (xx (length location))
                (let ((point (elt location xx)))
                  (dotimes (k 2) (setf (elt point k) (funcall 
                    (ax-obj-transform-fn (assoc-val (elt axes k) gaxes))
                         (elt point k)))
                    (dbg :xform "axis is ~A" (elt axes k))
                ))))
            (setf (gr-obj-coords gelem) 'absolute))
          ))))
;; ------------------------------------------------------ show
(defun show (gr-list &optional (options ()))
  "show (gr-list &optional (options ())) -- compile the graphics
  directives in glist and write the output to *graphics-terminal*"
  ;; process options. things like terminal, xsize, ysize, color, ...
  (let* ((goptions (alist-join options *graphics-options*))
         (gl    (copy-glist gr-list))
  ;; scan gl for bboxes in axis coords. Determine the 
  ;; minimum and maximum extent for each axis. This may involve multiple
  ;; axes in the case of plots with left and right axes.
         (gaxes (graphics-compile gl goptions)))
   (dbg :show "goptions ~A " goptions) 
   (dbg :show "gaxes ~A " gaxes) 
   (dolist (gaxis gaxes)
      (setf gl (glist-add gl (make-axis-glist (cdr gaxis) goptions))))
   (setf gl (glist-add gl (make-legend-glist gl)))
  ;; scan glist for objects specified in axis-coordinates. Scale the
  ;; dimensions of these appropriately. Also,
  ;; scale objects specified in relative coordinates (0 to 1 -> min to max)
   (dbg :show2 "gr-list before xform ~A " gr-list) 
   (graphics-transform gl gaxes goptions)
   (dbg :show2 "gr-list after xform ~A " gr-list) 
   (dbg :show "axes~A" gaxes)

  ;; paint the picture, starting with the first object in gl.
   (graphics-display gl goptions)
   gr-list
  ))
;; ------------------------------------------------------ open-terminal
(defun open-terminal ()
"open-terminal () -- This routine is clisp/linux specific. It opens
the postscript viewer. returns *gv-pid*"
#+(and clisp unix)
  (unless *gv-pid* 
    (if (find :cygwin *features*) ; cygwin doesn't support ps -auwwx
      (let ((cmd                  ; create shell script to run gv in bkgnd.
        (format nil "~A ~A" *postscript-viewer* *graphics-terminal*)))
        (with-open-stream (s (ext:run-shell-command cmd :output :stream))
           (read s) (defparameter *gv-pid* (read s)))
       *gv-pid*)
    ; else
      (progn
        (ext:shell 
          (format nil "~A ~A &" *postscript-viewer* *graphics-terminal*))
        (let ((cmd 
          (format nil 
             "ps auwwx | grep ~A | grep -v grep | awk '// {print $2}'" 
             *graphics-terminal*)))
          (defparameter *gv-pid* 
            (with-open-stream (s (ext:run-shell-command cmd :output :stream))
              (read s)))))))
  )
;; ------------------------------------------------------ refresh-terminal
(defun refresh-terminal ()
"refresh-terminal () -- refresh the terminal. This works for gv by sending
a HUP signal to gv."
#+(and clisp unix)
  (unless (zerop (ext:shell (format nil "kill -s HUP ~D" *gv-pid*))) 
    (setf *gv-pid* nil) (open-terminal)))
;; ------------------------------------------------------ link-terminal
(defun link-terminal (pnum)
#+(and clisp unix)
 (let ((tfile (format nil "plot~4,'0D.eps" pnum)))
    (dbg :link-terminal "~%tfile ~A" tfile)
    (ext:shell (format nil "ln -fs ~A ~A" tfile 
                           (namestring *graphics-terminal*)))
    pnum))
;; ------------------------------------------------------ ffwd
(defun ffwd (&optional (n 1))
  (link-terminal (incf *current-plot* n))
  (refresh-terminal))
;; ------------------------------------------------------ rewind
(defun rewind (&optional (n 1)) (ffwd (- n)))
;; ------------------------------------------------------ show-current
(defun show-current () (link-terminal *plot-counter*) (refresh-terminal))
;; ------------------------------------------------------ graphics-display
(defun graphics-display (gl &optional (options ()))
  (dbg :display "goptions ~A " options) 
  (dbg :display "xsize ~A " (cdr (assoc 'xsize options :test #'(lambda (x y)
    (equal (symbol-name x) (symbol-name y))))))
  (dbg :glist "graphics-display: gl ~A" gl)
  (defparameter *current-plot* (link-terminal (incf *plot-counter*)))

  (write-prolog (assoc-val 'xsize options) (assoc-val 'ysize options))
  (with-open-file (gout *graphics-terminal* 
                        :direction :output :if-exists :append)
    (dolist (gelem (glist-lst gl))
      (case (gr-obj-type gelem)
        ((axis-setting legend-entry) ()) ; these have already been "displayed"
        (otherwise
          (when (gr-obj-display-fn gelem)
            (funcall (gr-obj-display-fn gelem) gout gelem)))))
    (format gout "~&%%EOF")
    )
  (open-terminal)
  (show-current)
  )
;;
;;  This program is free software; you can redistribute it and/or modify
;;  it under the terms of the GNU General Public License as published by
;;  the Free Software Foundation; either version 2 of the License, or
;;  (at your option) any later version.
;;
;;  This program is distributed in the hope that it will be useful,
;;  but WITHOUT ANY WARRANTY; without even the implied warranty of
;;  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;  GNU General Public License for more details.
;;
;;  You should have received a copy of the GNU General Public License
;;  along with this program; if not, write to the Free Software
;;  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
;;
;;  Contact: Steve Eichblatt
;;  620 E San Salvador San Jose, CA 95112. steve@eichblatt.us


