;;  plot.lsp. The Data Touring Machine. Copyright (C) 2004 Steve Eichblatt
;;
;;  This program is free software; you can redistribute it and/or modify
;;  it under the terms of the GNU General Public License as published by
;;  the Free Software Foundation; either version 2 of the License, or
;;  (at your option) any later version.
;;
;;  This program is distributed in the hope that it will be useful,
;;  but WITHOUT ANY WARRANTY; without even the implied warranty of
;;  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;  GNU General Public License for more details.
;;
;;  You should have received a copy of the GNU General Public License
;;  along with this program; if not, write to the Free Software
;;  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
;;
;;  Contact: Steve Eichblatt
;;  620 E San Salvador San Jose, CA 95112. steve@eichblatt.us

(print (find-package :graphics) )
(unless (find-package :utils)    (load (compile-file "./utils")))
(unless (find-package :graphics) (load (compile-file "./graphics")))
(unless (find-package :color)    (load (compile-file "./color")))
(unless (find-package :graphics) (print "didn't load graphics right"))
(defpackage :datoura-plot
           (:nicknames :plot)
           (:use :graphics :common-lisp :utils)
           (:export #:profile #:scatter-plot #:histogram #:bar-graph
                    #:fun-plot #:parametric-plot #:xyz-plot))
(in-package :datoura-plot)

(defparameter *scatter-plot-options* '((join . ())))
(defparameter *fun-plot-options* '((join . t)))
(defparameter *xyz-plot-options* '(()))
(defparameter *parametric-plot-options* '((join . t)))
(defparameter *histogram-options* '())
(defparameter *profile-options* '())
(defparameter *bar-graph-options* '())
;; ------------------------------------------------------ bar-graph
(defun bar-graph (seq &key (barwidth .95) (offset 0) name
                           fill color options)
"bar-graph (seq &key (barwidth .95) (offset 0) &options)
  return a list of graphics objects describing the bar-graph
  seq is of the form (('x1' yval) ('x2' yval) ...)
  Needs to do some sanity-checking!!!
  Options include min max binwidth xlabel ylabel title 
  (needs: ymax logy color fill linetype)"
  (let* ((poptions (alist-join options *bar-graph-options*))
        (data (kill-nil seq 1))
        (xseq (map 'list #'princ-to-string 
                      (map 'list #'elt-0 data)))
        (xdat (range (1- (length xseq))))
        (xaxis nil) (yaxis nil)
        (yseq (map 'list #'elt-1 data))
        (marker-label-fn #'(lambda (x xx) (declare (ignore xx))
           (if (< x (length xseq)) (elt xseq x) "")))
        (gl (make-glist))) ;( (binwidth 1))
    (dbg :bar-graph "poptions ~A" poptions)
    (dbg :bar-graph "y data ~A" yseq)
    (multiple-value-setq (gl xaxis yaxis) 
        (make-plot-axes -.5 0 (- (length xdat) .5) 0 poptions))
    (when color (glist-add gl (set-color color)))
    (dotimes (k (length xdat))
      (glist-add gl
        (rectangle `((,(+ (- (elt xdat k) (/ barwidth 2)) offset) 0) 
                         (,barwidth ,(elt yseq k)))
        `((axes . (,xaxis ,yaxis))(rec-type . ,(when fill "rectfill"))))  
        ))
    (when color (glist-add gl (unset-color)))
    (when name (glist-add gl 
      (legend-entry name :symbol :lgsq :rgbcolor color)))
    (glist-add gl
      (axis-setting xaxis :major-ticks 1 :n-minor-ticks 0 
                                :marker-label-fn marker-label-fn))
    (values gl xdat yseq)
))
;; ------------------------------------------------------ fun-plot
(defun fun-plot (func xmin xmax &key dx (npoints 200) ymin ymax 
   (join t) (symbol :smpt) color options)
  "return a list of graphics objects describing the function plot.
  Options include ..."
  (let* ((poptions (alist-join options *fun-plot-options*))
        (dx (or dx (/ (- xmax xmin) npoints)))
        (domain (range xmin xmax dx))
        (vals (mapcar func domain)))
     (scatter-plot (thread domain vals) :ymin ymin :ymax ymax
        :join join :symbol symbol :color color :options poptions)))
;; ------------------------------------------------------ histogram
(defun histogram (seq &key binwidth fill color name normalize 
                      stats line dash options)
  "return a list of graphics objects describing the histogram
  Options include min max binwidth xlabel ylabel title 
  (needs: ymax logx logy axis color fill linetype)"
  (let* ((poptions (alist-join options *histogram-options*))
        (xstats (simple-stats seq)) (xaxis nil) (yaxis nil)
        (xmin (or (assoc-val 'min poptions) (assoc-val 'min xstats))) 
        (xmax (or (assoc-val 'max poptions) (assoc-val 'max xstats)))
        (data (remove-if-not #'numberp seq)) (frequency ())
        (gl (make-glist)))
    (dbg :histogram "poptions ~A" poptions)
    (dbg :histogram "binwidth ~A min ~A max ~A" binwidth xmin xmax)
    (setf data (remove-if #'(lambda (x) (or (< x xmin) (> x xmax))) data))
    (unless binwidth (setf binwidth (/ (calc-major-ticks xmin xmax) 2)))
    (setf frequency (frequencies data binwidth 
                     :normalize normalize))
    (dbg :histogram "binned data ~A" data)
    (dbg :histogram "frequency ~A" frequency)
    (multiple-value-setq (gl xaxis yaxis) 
        (make-plot-axes xmin 0 xmax 0 poptions))
    (when color (glist-add gl (set-color color))) 
    (case (or dash line)
       ((nil)
       (dolist (elem frequency)
        (glist-add gl
          (rectangle (list (list (first elem) 0) (list binwidth (second elem)))
          `((axes . (,xaxis ,yaxis))(rec-type . ,(when fill "rectfill"))))  
          )))
       ((t)
        (glist-add gl (ps-directive "gsave 2 setlinewidth"))
        (when dash (glist-add gl (ps-directive "[3 5] 6 setdash")))
        (glist-add gl (line 
           (mapcar #'(lambda (x) (list (+ (first x) (/ binwidth 2)) (second x)))
                        frequency) 
           `(axes . (,xaxis ,yaxis)))) 
        (glist-add gl (grestore)))
      (otherwise ()))
    (when color (glist-add gl (unset-color)))   
    (if stats
      (glist-add gl (legend-entry 
          (format nil "~A N:~5D Mean:~8,3F RMS:~7,3F" (or name "")
                       (assoc-val :n xstats)(assoc-val :mean xstats)
                       (assoc-val :rms xstats))
          :rgbcolor (or color '(0 0 0))))
      (when name (glist-add gl 
         (legend-entry name :symbol :lgsq :rgbcolor color))))
    (values gl frequency xstats)))
;; ------------------------------------------------------ make-plot-axes
(defun make-plot-axes (xmin ymin xmax ymax &optional (options ()))
"make-plot-axes (xmin ymin xmax ymax &optional (options ())) -- makes 
the plot axes and returns the glist"
  (let ((gl ())
        (xaxis (or (assoc-val 'xaxis options) :x0))
        (yaxis (or (assoc-val 'yaxis options) :y0)))
    (when (and xmin ymin)
      (glist-add gl (stretch-point xmin ymin 
                        `(axes . (,xaxis ,yaxis)))))
    (when (and xmax ymax)
      (glist-add gl (stretch-point xmax ymax 
                       `(axes . (,xaxis ,yaxis)))))
    (glist-add gl 
      (axis-setting xaxis :label (assoc-val 'xlabel options)
                 :major-ticks   (assoc-val 'x-major-ticks options)
                 :n-minor-ticks (assoc-val 'x-minor-ticks options)))
    (glist-add gl 
      (axis-setting yaxis :label (assoc-val 'ylabel options)
                 :major-ticks   (assoc-val 'y-major-ticks options)
                 :n-minor-ticks (assoc-val 'y-minor-ticks options)))
    (unless (equal yaxis :y0) 
      (glist-add gl (axis-setting yaxis :location .85)))
    (glist-add gl  
      (text .5 .95 (assoc-val 'title options) 
         '((text-justification . center)(text-font . "times-roman-large")
           (coords . relative))))
    (values gl xaxis yaxis)))
;; ------------------------------------------------------ maxamillion
(defun maxamillion (seq xrng yrng &optional (n 300))
"maxamillion (seq xrng yrng &optional (n 300)) -- 
to prevent scatter plots from getting too big
called 'maxamillion' since it limits number of points (optionally) 
to a million"
  (let ((xgrid (/ xrng n)) (ygrid (/ yrng n)))
    (remove-duplicates 
       (map 'vector 
         #'(lambda (x) (round-to-nearest-point x xgrid ygrid)) seq) 
          :test #'equal)))
;; ------------------------------------------------------ parametric-plot
(defun parametric-plot (xfunc yfunc tlist &key xmin xmax ymin ymax 
   (join t) (symbol :smpt) options)
  "return a list of graphics objects describing the parametric plot.
  Options include ..."
  (let* ((poptions (alist-join options *parametric-plot-options*)))
     (scatter-plot (thread (mapcar xfunc tlist) (mapcar yfunc tlist)) 
        :xmin xmin :xmax xmax :ymin ymin :ymax ymax
        :join join :symbol symbol :options poptions)))

;; ------------------------------------------------------ profile
(defun profile (seq &key binwidth (offset 0) name stats
                         name xmin xmax ymin ymax color options)
  "return a list of graphics objects describing the profile.
  seq is of the form: ((x1 y1) (x2 y2) ...)
  Options include ..."
  (declare (optimize (safety 3)))
  (let* ((poptions (alist-join options *scatter-plot-options*))
        (seq (kill-nil seq 1))
        (xseq (map 'vector #'elt-0 seq))
        (yseq (map 'vector #'elt-1 seq))
        (xstats (simple-stats xseq)) (ystats (simple-stats yseq))
        (gl (make-glist)) (means ()) (errbars ())
        (xn  (make-hash-table :test #'equal)) 
        (xs  (make-hash-table :test #'equal)) 
        (xs2 (make-hash-table :test #'equal)) 
        (yn  (make-hash-table :test #'equal)) 
        (ys  (make-hash-table :test #'equal)) 
        (ys2 (make-hash-table :test #'equal)))
    (when (or xmin xmax ymin ymax)
      (setf seq (remove-if 
                #'(lambda (p) 
                    (or (< (elt p 0) xmin) (> (elt p 0) xmax)
                        (< (elt p 1) ymin) (> (elt p 1) ymax))) seq))
      (setf xseq (map 'vector #'elt-0 seq))
      (setf yseq (map 'vector #'elt-1 seq)))
    (dbg :profile "xseq ~A xstats ~A" xseq xstats)             
    (setf xmin (or xmin (assoc-val :min xstats)))
    (setf xmax (or xmax (assoc-val :max xstats)))
    (setf ymin (or ymin (assoc-val :min ystats)))
    (setf ymax (or ymax (assoc-val :max ystats)))
    (setf binwidth (or binwidth (/ (calc-major-ticks xmin xmax) 2)))
    (dbg :profile "xmin ~A xmax ~A binwidth ~A" xmin xmax binwidth)             

    (dotimes (k (length seq))
      (let ((x (elt xseq k)) (y (elt yseq k))
            (key ()))
        (setf key (format nil "~8,4E" (floor-to-nearest x binwidth)))
        (incf (gethash key xn  0))
        (incf (gethash key yn  0))
        (incf (gethash key xs  0) x)
        (incf (gethash key ys  0) y)
        (incf (gethash key xs2 0) (* x x))
        (incf (gethash key ys2 0) (* y y))))
    (dbg :profile "color ~A" color)
    (dolist (key (sort (keys xn) 
                    #'(lambda (x y) 
                      (< (read-from-string x)(read-from-string y)))))
      (let* ((xnv  (gethash key xn  0)) (ynv  (gethash key yn  0))
             (xsv  (gethash key xs  0)) (ysv  (gethash key ys  0))
             (xs2v (gethash key xs2 0)) (ys2v (gethash key ys2 0))
             (xmean (/ xsv xnv))
             (ymean (/ ysv ynv))
             (xrms  (sqrt (- (/ xs2v xnv)(* xmean xmean))))
             (yrms  (sqrt (- (/ ys2v ynv)(* ymean ymean))))
             (rootN (sqrt xnv))
             (ylo (- ymean (/ yrms rootN))) (yhi (+ ymean (/ yrms rootN))))
         (push `(,xmean ,ymean) means)
         (push `((,xmean ,yhi)(,xmean ,ylo)) errbars)))
    (when color (glist-add gl (set-color color)))
    (glist-add gl (list (ps-directive "gsave 2 setlinewidth")))
    (glist-add gl (scatter-plot means :join t :options poptions))
    (glist-add gl (scatter-plot means :symbol :lgpt :options poptions))
    (dbg :profile "errbars ~A" errbars)             
    (dolist (elem errbars)
      (glist-add gl (scatter-plot elem :join t )))  ; options poptions
    (glist-add gl (ps-directive "grestore"))
    (if stats
      (glist-add gl (legend-entry 
          (format nil 
            "~A N:~5D Yave:~8,3F Xave:~8,3F"
              (or name "") (assoc-val :n xstats)
              (assoc-val :mean ystats) (assoc-val :mean xstats)) 
          :rgbcolor (or color '(0 0 0))))
      (when name (glist-add gl 
         (legend-entry name :symbol :lgsq :rgbcolor color))))
    (when color (glist-add gl (unset-color)))
    (values gl xstats ystats means errbars)
 )) 
;; ------------------------------------------------------ scatter-plot
(defun scatter-plot (seq &key color xmin xmax ymin ymax join
   (symbol :mdpt) name stats profile profile-color xform xticks options)
  "return a list of graphics objects describing the scatter plot.
  seq is of the form: ((x1 y1) (x2 y2) ...)
  Options include ..."
  (let ((poptions (alist-join options *scatter-plot-options*))
        (seq (kill-nil seq 1))
        (xstats (simple-stats (map 'vector #'elt-0 seq)))
        (ystats (simple-stats (map 'vector #'elt-1 seq)))
        (gl (make-glist)) (xaxis nil) (yaxis nil))
    (setf xmin (or xmin (assoc-val 'min xstats)))
    (setf xmax (or xmax (assoc-val 'max xstats)))
    (setf ymin (or ymin (assoc-val 'min ystats))) 
    (setf ymax (or ymax (assoc-val 'max ystats)))
    (setf seq (remove-if 
                #'(lambda (p) 
                    (or (< (elt-0 p) xmin) (> (elt-0 p) xmax)
                        (< (elt-1 p) ymin) (> (elt-1 p) ymax))) seq))
    (dbg :scatter-plot "xmin ~A xmax ~A ymin ~A ymax ~A" xmin xmax ymin ymax)
    (setf xmin (floor-to-nearest xmin (calc-major-ticks xmin xmax)))
    (setf ymin (floor-to-nearest ymin (calc-major-ticks ymin ymax)))
    (setf xmax (ceiling-to-nearest xmax (calc-major-ticks xmin xmax)))
    (setf ymax (ceiling-to-nearest ymax (calc-major-ticks ymin ymax)))
    (dbg :scatter-plot "xmin ~A xmax ~A ymin ~A ymax ~A" xmin xmax ymin ymax)
    (when (> (length seq) 10000) 
       (setf seq (maxamillion seq (- xmax xmin) (- ymax ymin))))
    (multiple-value-setq (gl xaxis yaxis) 
        (make-plot-axes xmin ymin xmax ymax poptions))
    (when color (glist-add gl (set-color color)))
    (glist-add gl
      (cond (join (line seq `(axes . (,xaxis ,yaxis)))) 
            (t (map 'list #'(lambda (x) 
                              (point (elt-0 x) (elt-1 x) 
                             `((point-symbol . ,symbol)
                              (axes . (,xaxis ,yaxis))))) 
                  seq))))
    (when color (glist-add gl (unset-color)))
    (when profile 
      (when profile-color (setf color profile-color))
      (if (and (not color) (not profile-color)) 
         (setf color color::red)
         (setf color (mapcar #'(lambda (x) (* .8 x)) color)))
      (glist-add gl
          (profile seq :color color :options (acons 'title nil poptions))))
    (when xform (glist-add gl 
       (axis-setting xaxis :marker-label-fn xform)))
    (when xticks (glist-add gl 
       (axis-setting xaxis :major-ticks xticks)))
    (if stats
      (glist-add gl (legend-entry 
          (format nil 
            "~A N:~5D Yave:~8,3F Xave:~8,3F"
              (or name "") (assoc-val :n xstats)
              (assoc-val :mean ystats) (assoc-val :mean xstats))
          :symbol symbol :rgbcolor (or color '(0 0 0))))
      (when name (glist-add gl 
         (legend-entry name :symbol symbol :rgbcolor (or color '(0 0 0))))))
    (values gl xstats ystats)
    )) 

(defun zcolor (zval zmin zmax)
"zcolor (zval zmin zmax)
  return a rgb color based on the zval between zmin (red) and zmax (green)"
  (let* ((zrange (- zmax zmin))
         (z2norm (* 2 (- (/ (- zval zmin) zrange) .5)))
         (redval   (if (< z2norm 0) (abs z2norm) 0))
         (greenval (if (> z2norm 0) (abs z2norm) 0)))
    (list redval greenval 0)))

;; ------------------------------------------------------ xyz-plot
(defun xyz-plot (seq &key xwidth ywidth name
                          zmin zmax options)
"xyz-plot (seq &key (xwidth .01) (ywidth .01) name zmin zmax options)
  return a list of graphics objects describing the xyz-plot 
  seq is of the form ((x1 y1 z1) (x2 y2 z2) ...) "
  (let* ((poptions (alist-join options *xyz-plot-options*))
        (data (kill-nil seq 1))
        (xaxis nil) (yaxis nil)
        (xseq (map 'list #'elt-0 data))
        (yseq (map 'list #'elt-1 data))
        (zseq (map 'list #'elt-2 data))
        (xstats (simple-stats xseq))
        (ystats (simple-stats yseq))
        (zstats (simple-stats zseq))
        (gl (make-glist)) (xaxis nil) (yaxis nil))

    (setf xmin (assoc-val 'min xstats)) 
    (setf xmax (assoc-val 'max xstats))
    (setf ymin (assoc-val 'min ystats)) 
    (setf ymax (assoc-val 'max ystats))
    (setf zmin (or zmin (assoc-val 'min zstats))) 
    (setf zmax (or zmax (assoc-val 'max zstats)))
    (setf xwidth (or xwidth (/ (- xmax xmin) 100)))
    (setf ywidth (or ywidth (/ (- ymax ymin) 100)))
    (dbg :xyz-plot "poptions ~A" poptions)
    (dbg :xyz-plot "y data ~A" yseq)
    (multiple-value-setq (gl xaxis yaxis) 
        (make-plot-axes xmin ymin xmax ymax poptions))
    (dotimes (k (length xseq))
      (glist-add gl (set-color (zcolor (elt zseq k) zmin zmax)))
      (glist-add gl
        (rectangle `((,(- (elt xseq k) (/ xwidth 2))
                      ,(- (elt yseq k)(/ ywidth 2))) 
                         (,xwidth ,ywidth))
        `((axes . (,xaxis ,yaxis))(rec-type . "rectfill"))))  
      (glist-add gl (unset-color)))
    (glist-add gl (set-color (zcolor zmax zmin zmax)))
    (glist-add gl (text .9 .8 (format nil "~6,3F" zmax) 
          '((coords . relative))))
    (glist-add gl (unset-color))
    (glist-add gl (set-color (zcolor zmin zmin zmax)))
    (glist-add gl (text .9 .75 (format nil "~6,3F" zmin) 
          '((coords . relative))))
    (glist-add gl (unset-color))
    (glist-add gl  
      (text .5 .95 (assoc-val 'title options) 
         '((text-justification . center)(text-font . "times-roman-large")
           (coords . relative))))
    (when name (glist-add gl 
      (legend-entry name :symbol :lgsq :rgbcolor color)))
    (values gl xseq yseq)
))

