;;  utils.lsp. The Data Touring Machine. Copyright (C) 2004 Steve Eichblatt
;;
;;  This program is free software; you can redistribute it and/or modify
;;  it under the terms of the GNU General Public License as published by
;;  the Free Software Foundation; either version 2 of the License, or
;;  (at your option) any later version.
;;
;;  This program is distributed in the hope that it will be useful,
;;  but WITHOUT ANY WARRANTY; without even the implied warranty of
;;  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;  GNU General Public License for more details.
;;
;;  You should have received a copy of the GNU General Public License
;;  along with this program; if not, write to the Free Software
;;  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
;;
;;  Contact: Steve Eichblatt
;;  620 E San Salvador San Jose, CA 95112. steve@eichblatt.us

(defpackage :datoura-utils
            (:nicknames :utils)
            (:use :common-lisp)
            (:export #:assoc-val #:alist-join #:degtorad #:positions 
               #:stretch-list #:mac #:mapstretch #:mapstretch-2 #:commap
               #:thread-lists #:thread #:type-seq #:extract #:? 
               #:flatten #:depth #:debug-on #:debug-off #:dbg
               #:*dbg-ids* #:elt-0 #:elt-1 #:elt-2 #:loadhash
               #:delay #:force #:mmddyy #:mmyy #:mv-2 #:mv-n
               #:intersect-svec #:set-diff-svec #:position-svec
               #:defun-memo #:memo #:memoize #:clear-memoize
               #:keys #:print-keys #:print-hash #:kill-nil
               #:floor-to-nearest #:ceiling-to-nearest
               #:deal-random #:uniqify #:rename-duplicates
               #:make-random-list #:round-to-nearest #:name-of
               #:round-to-nearest-point #:arguments #:hash-project
               #:make-hash-pipe #:gethash-pipe #:hash-pipe-hash
               #:range #:split-string #:in #:sort-copy #:sort-in-place
               #:list-order #:nth-or-last #:asc-desc-p #:grouplist
               #:cdf #:frequencies #:category-freq #:abbrev #:abbrevs
               #:list-sort-asc-desc #:list-asc-p #:simple-stats))
(in-package :datoura-utils)
;; -------------------------------------------------------- ?
(defun ? (fun)
  "? (fun) -- print the documentation string for function fun"
  (documentation fun 'function))
;; --------------------------------------------------------- abbrev(s)
(defmacro abbrev (short long)
"abbrev (short long) -- from Paul Graham's On Lisp, p 214"
  `(defmacro ,short (&rest args)
    `(,',long ,@args)))
(defmacro abbrevs (&rest names)
"abbrevs (&rest names) -- from Paul Graham's On Lisp, p 214"
  `(progn
     ,@(mapcar #'(lambda (pair) `(abbrev ,@pair)) (grouplist names 2))))
;; ------------------------------------------------------ alist-join
(defun alist-join (alist blist)
  "alist-join (alist blist) -- prepend assoc list a to assoc list b. 
  If alist is a single element like '(q . r) then cons it to the list,
  otherwise append blist to alist"
  (cond ((null alist) blist)
        ((= 1 (depth alist)) (cons alist blist))
         (t (append alist blist))))
;; ------------------------------------------------------ arguments
(defun arguments (lis) 
"Return arguements of lists in a tree. For example 
(arguments '((+ a b) c (* (+ (sin x) (cos y)) 2))) 
returns (a b c x y 2)"
(cond ((null lis) nil)
      ((atom (car lis)) (cons (car lis) (arguments (cdr lis))))
      ((listp (car lis))  (arguments (append (cdr (car lis)) (cdr lis))))
));; -------------------------------------------------------- asc-desc-p
(defun asc-desc-p (a b fnlist)
  "asc-desc-p (a b fnlist) -- predicate for sorting lists by 
  ascending or descending columns, controlled by fnlist.
  Also -- handle character comparison!"
  ; fnlist can be shorter than a b
  (setf fnlist (append fnlist (last fnlist))) 
  (cond ((null a) t)
        ((null b) nil)
        ((funcall (first fnlist) (or (elt a 0) -999) (or (elt b 0) -999)) t)
        ((equal (elt a 0) (elt b 0)) 
          (asc-desc-p (subseq a 1) (subseq b 1) (rest fnlist)))
        (t nil)))
;; --------------------------------------------------------- assoc-val 
(defun assoc-val (key alist)
  "assoc-val (key alist) -- retrieve the value of the alist with key,
  instead of the pair. return nil if key not in alist"
  (dbg :assoc-val "key ~A Alist ~A " key alist)
  (cdr (assoc key alist :test #'(lambda (x y)
    (equal (symbol-name x) (symbol-name y))))))
;; ------------------------------------------------------ category-freq
(defun category-freq (seq)
"category-freq (seq) -- categorical frequencies. "
  (let* ((categories (remove-duplicates seq)))
    (setf categories (sort categories #'string-lessp))
    (map 'list #'(lambda (bin) (list bin (count bin seq))) categories)))
;; --------------------------------------------------------- cdf
(defun cdf (seq &key normalize)
"cdf (seq &key normalize) -- return a list of pairs of x y values
describing the cumulative distribution of the sequence."
  (let* ((len (length seq))
         (y   (if normalize (range (/ normalize len) normalize (/ normalize len))
                            (range 1 (+ len 1))))
         (data (remove-if-not #'numberp seq)))
    (setf data (sort data #'<))
    (thread data y))) 
;; ------------------------------------------------------ ceiling-to-nearest
(defun ceiling-to-nearest (num precision)
 "ceiling-to-nearest (num precision) -- 
  round num up to the next value evenly divisible by precision. 
  Works best if precision is a rational!"
  (* (ceiling num precision) precision))
;; ----------------------------------------------------- debug stuff
;;; Code from Paradigms of AI Programming
;;; Copyright (c) 1991 Peter Norvig
;;;; The Debugging Output Facility:
(defvar *dbg-ids* nil "Identifiers used by dbg")
(defun dbg (id format-string &rest args)
  "Print debugging info if (DEBUG ID) has been specified."
  (when (member id *dbg-ids*)
    (fresh-line *debug-io*)
    (format *debug-io* "~A: " id)    ;; added SE 6/6/04
    (apply #'format *debug-io* format-string args)))
(defun debug-on (&rest ids)
  "Start dbg output on the given ids."
  (setf *dbg-ids* (union ids *dbg-ids*)))
(defun debug-off (&rest ids)
  "Stop dbg on the ids.  With no ids, stop dbg altogether."
  (setf *dbg-ids* (if (null ids) nil
                      (set-difference *dbg-ids* ids))))
;;; ==============================
(defun dbg-indent (id indent format-string &rest args)
  "Print indented debugging info if (DEBUG ID) has been specified."
  (when (member id *dbg-ids*)
    (fresh-line *debug-io*)
    (dotimes (i indent) (princ "  " *debug-io*))
    (apply #'format *debug-io* format-string args)))
;; -------------------------------------------------------- mmddyy
(defun mmddyy (datetime &optional xx) 
"ddmmyy (datetime &optional xx) -- convert datetime number to string" 
(declare (ignore xx)) 
(let ((l (multiple-value-list (decode-universal-time datetime))))
  (format nil "~A/~A/~2,'0D" (nth 4 l) (nth 3 l) (mod (nth 5 l) 100))))
;; -------------------------------------------------------- mmyy
(defun mmyy (datetime &optional xx) 
"ddmmyy (datetime &optional xx) -- convert datetime number to string" 
(declare (ignore xx)) 
(let ((l (multiple-value-list (decode-universal-time datetime))))
  (format nil "~A/~2,'0D" (nth 4 l) (mod (nth 5 l) 100))))
;; -------------------------------------------------------- degtorad
(defun degtorad (deg) "degtorad (deg) -- convert degrees to radians" 
(* deg (/ pi 180)))
;; -------------------------------------------------------- depth
(defun depth (tree &optional (d0 1))
"depth (tree &optional (d0 0)) -- return depth of deepest node in tree"
  (cond ((null tree) d0)
        ((atom tree) d0)  ; handles alists.
        ((atom (first tree)) (depth (rest tree) d0))
        (t (max (depth (first tree) (+ 1 d0)) (depth (rest tree) d0 )))))
;; --------------------------------------------------------- elt-0
(defun elt-0 (seq)
 "Sequence analog to list function first"
 (elt seq 0))
;; --------------------------------------------------------- elt-1
(defun elt-1 (seq)
 "Sequence analog to list function second"
 (elt seq 1))
;; --------------------------------------------------------- elt-2
(defun elt-2 (seq)
 "Sequence analog to list function third"
 (elt seq 2))
;; --------------------------------------------------------- extract
(defun extract (seq elems)
" extract (lis elems) -- Return the elements of seq which appear
  in the sequence elems"
  (if (null seq) ()
    (map (type-seq seq) #'(lambda (n) (elt seq n)) elems)))
;; -------------------------------------------------------- flatten
(defun flatten (tree &optional (level 0))
"flatten (tree) -- modified from Essential Lisp, Anderson,Corbett, 
  Reiser p. 149. Added level option -- i'm not sure this really does
  what it should...should it flatten from top or bottom?"
  (cond ((null tree) nil)
        ((zerop level) (flatten-full tree))
        ((atom (first tree)) (cons (first tree) (flatten (rest tree) level)))
        ((not (> (depth (first tree)) level)) 
          (cons (first tree) (flatten (rest tree) level)))
        (t (append (flatten (first tree) level) 
                   (flatten (rest tree ) level)))))
;; -------------------------------------------------------- flatten-full
(defun flatten-full (tree)
"flatten-full (tree) -- taken from On Lisp, Graham, p390."
   (mapcan #'(lambda (x) 
       (if (atom x) (mklist x) (flatten-full x)))
   tree))
;; 00------------------------------------------------------ floor-to-nearest
(defun floor-to-nearest (num precision)
 "floor-to-nearest (num precision) -- 
 round num down to the next value evenly divisible by precision. 
 Works best if precision is a rational!"
  (* (floor num precision) precision))
;; -------------------------------------------------------- frequencies
(defun frequencies (seq binwidth &key normalize)
 "frequencies (seq binwidth &key normalized) -- 
 return a list of pairs of (binedge, count) pairs based on data seq. 
 If normalized, then return pairs of (binedge, probability). It would be 
 nice to have this work for
 character sequences also"
  (let* ((data (map (type-seq seq)  
                  #'(lambda (x) (floor-to-nearest x binwidth)) 
               (remove-if-not #'numberp seq)))
         (bindata (remove-duplicates data)))
    (setf bindata (sort bindata #'<))
    (setf bindata (range (elt-0 bindata) 
                         (elt bindata (1- (length bindata))) binwidth))
    (setf bindata (map (type-seq bindata) 
         #'(lambda (x) (round-to-nearest x binwidth)) bindata))
    (map 'list #'(lambda (bin) 
       (list bin 
       (if normalize (* normalize (/ (count bin data) (length seq)))
                     (count bin data)))) 
    bindata)))
;; --------------------------------------------------------- grouplist
(defun grouplist (source n)
 "grouplist (source n) -- 
 from Paul Graham's group, On Lisp, p 47 return source list in 
 sublists of length n"
  (when (zerop n) (error "zero length")) 
  (labels ((rec (source acc)
             (let ((rest (nthcdr n source)))
               (if (consp rest)
                   (rec rest (cons (subseq source 0 n) acc))
                   (nreverse (cons source acc))))))
  (if source (rec source nil) nil)))
;; --------------------------------------------------------- hash-project
(defmacro hash-project (hash expr &key names)
"hash-project (hash expr &key names) -- 
Return a hash which is the projection of expr of the columns of input hash
eg. (hash-project h (a b (+ a b))) gives a hash with 3 keys a, b, and (+ a b)
Note: should i put the loop inside the map and only pass through the data once?"
  (let* ((keys (keys (eval hash)))
         (args (intersection keys (arguments expr)))
         (llist (thread args
           (loop for k in args
          collect `(gethash ',k ,hash)))))
  `(let ,llist 
       (loadhash (thread (or ,names ',expr)
         (list ,@(loop for x in expr collect
          `(map 'vector #'(lambda ,args (and ,@args ,x)) ,@args))))))))
;; --------------------------------------------------- hash-pipe
(defstruct hash-pipe
  "A hash data structure with a maximum number of keys.  Attempts 
  to put more than <depth> keys into the hash-pipe delete the 
  oldest key. A key's age is reset every time it is accessed." 
  hash         ; the data
 (depth 500)   ; how many items to store
  keys         ; the list of keys in order of latest use.
)
;; ---------------------------------------------- hash-pipe-reload
(defun hash-pipe-reload (key name)
"Set <name>'s most recently accessed key to <key> argument."
 (let ((keys (hash-pipe-keys name))
       (depth (hash-pipe-depth name))
       (hash (hash-pipe-hash name)))
   (setf keys (delete-duplicates (reverse (push key keys))
        :test (hash-table-test hash)))
   (when (> (length keys) depth)
      (remhash (pop keys) hash))
   (setf (hash-pipe-keys name) (nreverse keys))))
;; ---------------------------------------------- gethash-pipe
(defun gethash-pipe (key name)
"gethash-pipe (key name)
 hash-pipe's version of gethash"
  (let ((hash  (hash-pipe-hash name)))
     (multiple-value-bind (value found-p) (gethash key hash)
       (when found-p (progn (hash-pipe-reload key name)) 
       (values value found-p)))))
;; ---------------------------------------------- hash-pipe-update
(defun hash-pipe-update (key name value)
"hash-pipe-update (key name value)
 Internal function used by defsetf"
  (gethash-pipe key name)
  (hash-pipe-reload key name)
  (setf (gethash key (hash-pipe-hash name)) value))

(defsetf gethash-pipe hash-pipe-update)
;; -------------------------------------------------------- in
(defmacro in (obj &rest choices)
 "in (obj &rest choices) -- return t if obj is in choices, else ()
 From Graham's 'Common Lisp'"
  (let ((insym (gensym)))
    `(let ((,insym ,obj))
       (or ,@(mapcar #'(lambda (c) `(eql ,insym ,c)) choices)))))
;; --------------------------------------------------------- position-svec
(defun position-svec (item vec &key (start 0) end)
"position-svec (item vec &key (start 0)) --
  find item in vec, assuming that vec is sorted"
  (cond ((>= start (or end (length vec))) nil)
        (t (let* ((endpt (or end (length vec)))
                  (midpt (+ start (floor (/ (- endpt start) 2))))
                  (val (elt vec midpt)))
             (cond ((= val item) midpt)
                   ((> val item) 
                       (position-svec item vec :start start :end (1- midpt)))
                   ((< val item)
                       (position-svec item vec :start (1+ midpt) :end endpt))
                   (t nil))))))
;; --------------------------------------------------------- intersect-svec
(defun intersect-svec (vlong vshort)
"intersect-svec (vlong vshort) --
  find the intersection of 2 sorted vectors"
  (let ((result ()) (startpos 0))
     (dotimes (k (length vshort))
       (let* ((elem (elt vshort k))
             (ppos (position-svec elem vlong :start startpos)))
        (when ppos (setf startpos ppos) (push elem result))))
    (nreverse result)))
;; -------------------------------------------------------- keys
(defun keys (h)
  "keys (h) -- return the keys of a hash"
  (let ((k ()))
  (maphash #'(lambda (key val) (declare (ignore val))
               (push key k)) h)
  (nreverse k)))
;; -------------------------------------------------------- kill-nil
(defun kill-nil (seq &optional (level 0))
  "kill-nil (seq) -- 
  return seq without nils at top or first level 
  and the number of nils removed. A recursive version could handle
  any depth, but i'm not in the mood to do it right now."
  (let ((k 
    (case level 
      ((0) (remove nil seq))
      ((1) (remove-if #'(lambda (x) (position nil x)) seq))
      (otherwise ()))))
    (values k (- (length seq) (length k)))))
;; -------------------------------------------------------- list-asc-p
(defun list-asc-p (a b)
  "list-asc-p (a b) -- 
  predicate to sort lists ascending element-by-element"
   (list-order a b #'<))
(defun list-desc-p (a b)
  "list-desc-p (a b) -- 
  predicate to sort lists descending element-by-element"
   (list-order a b #'>))
;; -------------------------------------------------------- list-order
(defun list-order (a b fn)
  "list-order (a b fn) -- 
  return true if list a is (whatever fn says) of list b, on an 
  element-by-element comparison."
  (cond ((null a) t)
        ((null b) nil)
        ((funcall fn (first a) (first b)) t)
        ((equal (first a) (first b)) 
          (list-order (rest a) (rest b) fn))
        (t nil)))
;; ----------------------------------------------- list-sort-asc-desc
(defun list-sort-asc-desc (seq fnlist &optional tmpl)
  "list-sort-asc-desc (seq fnlist &optional tmpl) -- 
  sort a sequence of lists by ascending or descending order.
  eg. (list-sort-asc-desc '(((1 4 3) (2 3 5) (1 3 2) (2 3 8)) '(< >))
  tmpl is the template, eg. '(1 's 1) will sort using < string-lessp <.
  If tmpl is not provided, use the first element in seq."
  ; use appropriate comparison operators eg. string-lessp, etc.
  ; this would be best implemented with rules a la mma.
  (let ((template (or tmpl (elt seq 0))))
    (dotimes (k (length template))
      (if (or (symbolp (elt template k)) (stringp (elt template k))) 
        (if (equalp (elt fnlist k) '>) 
          (setf (elt fnlist k) 'string-greaterp)
          (setf (elt fnlist k) 'string-lessp)))))
  (sort seq #'(lambda (a b) (asc-desc-p a b fnlist))))
;; --------------------------------------------------- loadhash
(defun loadhash (keyvals &key (test #'equal))
  (let ((hash (make-hash-table :test test)))
    (dolist (kv keyvals)
       (setf (gethash (first kv) hash) (second kv)))
   hash))
;; --------------------------------------------------------- longest-branch
(defun longest-branch (tree)
 "longest-branch (tree) -- 
 return the length of the longest (first level) branch in tree"
  (apply #'max (mapcar #'(lambda (x) 
      (cond ((atom x) 1) 
            (t (length x)))) tree)))
;; --------------------------------------------------------- longest-of-2
(defun longest-of-2 (x y)
 "longest-of-2 (x y) -- 
 return the length of the longest (first level) branch in tree"
  (let ((xlen (if (consp x) (length x) 1))
        (ylen (if (consp y) (length y) 1)))
    (max xlen ylen)))
;; --------------------------------------------------------- mac
(defmacro mac (expr)
  `(pprint (macroexpand ',expr)))
;; -------------------------------------------------------- deal-random
(defun deal-random (n handsize)
"deal-random (n handsize) -- return a random list of n elements from the 
 integers 0 to (- handsize 1), with no repetitions"
  (if (< n 1) (setf n (floor (* n handsize))))
  (let ((base (range handsize))
        (result nil))
    (loop for elem in (range (1- n))
      collect (pop (subseq base (random (1+ (decf handsize))))))
  ))
;; -------------------------------------------------------- make-random-list
;; note -- optional args not allowed if tail-recursion optimized.
;;  use helper function
(defun mkrandlist (n mx lis)
  (if (= n 0)
    lis
    (mkrandlist (- n 1) mx (cons (random mx) lis) )))
(defun make-random-list (n &optional (mx 1.0) (lis ()))
  "make-random-list n &optional (mx 1.0) (lis ()) 
   Generate a list of random numbers from 0 to mx."
  (mkrandlist n mx lis))
;; --------------------------------------------------------- mapstretch-2
(defmacro mapstretch-2 (fn lst1 lst2)
  (let ((mxlen `(longest-of-2 ,lst1 ,lst2)))
    `(mapcar ,fn (stretch-list ,lst1 ,mxlen) 
                 (stretch-list ,lst2 ,mxlen))))
;; --------------------------------------------------------- mapstretch
(defmacro mapstretch (fn &rest lsts)
  (let ((mxlen `(longest-branch ,lsts)))
    `(mapcar ,fn (stretch-list ,(first  lsts) ,mxlen) 
                 (stretch-list ,(second lsts) ,mxlen))))
;; -------------------------------------------------------- mklist
(defun mklist (obj)
  (if (listp obj) obj (list obj)))
;; -------------------------------------------------------- mv-n
(defmacro mv-n (arg n) `(elt (multiple-value-list ,arg) (- ,n 1)))
;; -------------------------------------------------------- mv-2
(defmacro mv-2 (arg) `(mv-n ,arg 2))
;; --------------------------------------------------------- name-of
(defun name-of (x y)
 "name-of (x y) -- 
 return true if x and y are symbols with the same name. 
 This is required when using symbols as options across packages."
    (equal (symbol-name x) (symbol-name y)))
;; -------------------------------------------------------- nth-or-last
(defun nth-or-last (seq n)
  "nth-or-last (seq n) -- 
  return the nth element of seq, or the last if n is past the end of seq"
  (elt seq (min (1- (length seq)) n)))
; ---------------------------------------------------------- positions
(defun positions (test seq)
 "positions (test seq) -- 
  return a list of all of the positions (at the top level) 
  in seq where test is true.  "
  (let ((newlist nil))
    (dotimes (k (length seq) (nreverse newlist))
       (if (ignore-errors (funcall test (elt seq k)))
           (push k newlist)))))
;; -------------------------------------------------------- print-hash
(defun print-hash (h &optional (str t))
  "print-hash (h) --  
  recursively print the key-value pairs of a hash.  Needs indentation 
  levels to separate "
    (fresh-line)
    (maphash #'(lambda (key val)
                  (format str "~A -> " key)
                  (if (hash-table-p val)
                    (print-hash val str)
                    (format str "~A~%" val)))
     h))
;; -------------------------------------------------------- print-keys
(defun print-keys (h)
  "print-keys (h) -- 
  print the keys of a hash"
  (fresh-line)
  (maphash #'(lambda (key val) (declare (ignore val))
               (format t "~A " key)) h))
;; -------------------------------------------------------- range
(defun range (n &optional (mx ()) (inc 1))
  "range n &optional (mx ()) (inc 1) -- 
  generate numbers spaced by inc from 0 to n (inclusive) or, if mx is 
  not nil, from n to mx (inclusive)"
   (cond ((null mx) ; go from 0 to n by inc 
           (do ((i 0 (+ i inc)) (y '() (push i y))) ((> i n) (nreverse y))))
         ((or (and (< mx n) (> inc 0)) ; error
              (and (> mx n) (< inc 0))) () )
         (t 
           (do ((i n (+ i inc)) (y '() (push i y))) ((> i mx) (nreverse y))))
   )) 
;; ------------------------------------------------------ rename-duplicates
(defun rename-duplicates (lis)
"rename-duplicates (lis) -- given a list of strings, ensure that the names
are all distinct -- far from perfect. "
  (dbg :rename-duplicates "in rename-duplicates")
  (loop for i from 1 to (1- (length lis)) do
     (let* ((elem (elt lis i)) (n 1) (before (subseq lis 0 i))
            (after (subseq lis (1+ i))))
       (when (find elem before :test #'string-equal)
          (do ((j 2 (+ j 1)))
             (nil)
             (cond ((find (format nil "~A_~d" elem j) 
                          (append before after) :test #'string-equal)
                      nil)
                    (t (setf (elt lis i) (format nil "~A_~d" elem j)) (return)))))))
  lis)
;; ------------------------------------------------------ round-to-nearest
(defun round-to-nearest (num precision)
 "round-to-nearest (num precision) -- 
 round num to the nearest value evenly divisible by precision. 
 Works best if precision is a rational!"
  (* (round num precision) precision))
;; -------------------------------------------------- round-to-nearest-point
(defun round-to-nearest-point (pair &optional (x 1/1000) (y 1/1000))
 "round-to-nearest-point (pair &optional (x 1/1000) (y 1/1000))
  Round a point to the nearest point in space divisible by x y."
  (list (round-to-nearest (elt-0 pair) x) (round-to-nearest (elt-1 pair) y)))
;; --------------------------------------------------------- set-diff-svec
(defun set-diff-svec (vlong vshort)
  "set-diff-svec (vlong short) 
   return sequence of elements in vlong that are NOT in vshort. Both 
   vectors must be sorted before calling."
  (let ((result (copy-seq vlong))
        (startpos 0))
    (dotimes (k (length vshort))
      (let ((ppos (position (elt vshort k) vlong :start startpos)))
        (when ppos (setf (elt result ppos) nil) (setf startpos ppos))))
    (kill-nil result)))
;; -------------------------------------------------------- simple-stats
(defun simple-stats (seq)
 "simple-stats (seq) -- 
  return an assoc-list of stats for all numerical 
  elemens in seq (n,mean,rms,min,max) (character elements. ?)
  note: the sigma returned here is the standard deviation as defined by
  sqrt( sum_i (1/n * (x_i - x_mean)^2))...some people use n-1 instead of n
  also -- i added features form handling strings, but this really will
  mess things up for mixed columns.  "
  (let ((sum 0) (ss 0) (mean nil) (rms nil) (mmin nil) (mmax nil)
        (Nnum 0) (Nstring 0))
     (dotimes (k (length seq))
        (let ((elem (elt seq k)))
           (cond 
              ((numberp elem) 
                 (incf Nnum) (incf sum elem) (incf ss (* elem elem))
                 (setf mmin (if (null mmin) elem (if (< elem mmin) elem mmin)))
                 (setf mmax (if (null mmax) elem (if (> elem mmax) elem mmax))))
              ((stringp elem) 
                 (incf Nstring) 
                 (setf mmin (if (null mmin) elem 
                     (if (string-lessp elem mmin) elem mmin)))
                 (setf mmax (if (null mmax) elem 
                     (if (string-greaterp elem mmax) elem mmax))))
              (t nil))))
     (cond ((plusp Nnum) (setf mean (/ sum Nnum)) 
                      (setf rms (sqrt (- (/ ss Nnum) (* mean mean)))))
           (t nil))
     (list (cons :N (+ Nstring Nnum)) (cons :Mean mean) (cons :RMS rms)
           (cons :Min mmin) (cons :Max mmax) (cons :Sum sum))))
;; -------------------------------------------------------- sort-copy
(defun sort-copy (seq test)
  "sort-copy (seq test) -- 
  return a sorted copy of seq, without destroying seq"
  (let ((cp (copy-seq seq)))
    (sort cp test)))
;; -------------------------------------------------------- sort-in-place
(defmacro sort-in-place (seq test)
  "sort-in-place (seq test) -- 
  sort seq in place."
   `(setf ,seq (sort-copy ,seq ,test)))

;; -------------------------------------------------------- split-string
(defun commap (c) (and (graphic-char-p c)(char= c #\, )))
(defun whitespacep (c) (and (graphic-char-p c)(char= c #\  )))
(defun tokens (str &optional (test #'whitespacep) (start 0))
  " Taken from Paul Graham's Common Lisp, 'tokens' p 67. 
    tokens (str &optional (test #'whitespacep) (start 0))
    split a string into tokens."
  (let ((p1 (position-if test str :start start)))
    (if p1
       (let ((p2 (position-if #'(lambda (c) (not (funcall test c))) 
                                  str :start p1)))
         (cons (subseq str p1 p2)
           (if p2 (tokens str test p2)
              nil)))
     nil)))
(defun split-string (str &optional (test #'whitespacep))
  " derived from 'tokens' 
    split-string (str &optional (test #'whitespacep))
    split a string into tokens."
  (let ((pos (positions test str)) (len (length str)) (res ()))
     (setf pos (remove-duplicates (flatten (list -1 pos len))))
     (dotimes (k (1- (length pos)))
       (push (subseq str (1+ (elt pos k)) (elt pos (1+ k))) res))
     (nreverse res)))
;; Note: I would prefer a split-string that took a list of 
;; split characters as the second argument, like in perl.
;; --------------------------------------------------------- stretch-list
(defun stretch-list (lis len)
  (unless (consp lis) (setf lis (list lis)))
  (let ((lislen (length lis)))
    (cond ((= lislen len) lis)
          ((> lislen len) (subseq lis 0 len))
          (t (append lis 
             (make-list (- len lislen) :initial-element (first (last lis))))))))
;; -------------------------------------------------------- thread-lists
(defun thread-lists (lists)
 "thread-lists (lists) -- a la mathematica...  "
  (apply #'map (type-seq (car lists)) #'list lists))
;; -------------------------------------------------------- thread
(defun thread (&rest cols)
 "thread (&rest cols) -- a la mathematica...  "
  (thread-lists cols))
;; -------------------------------------------------------- type-seq
(defun type-seq (seq)
  (typecase seq
    (cons 'list) (vector 'vector)))

;;; ==============================
;;;; The Memoization facility:
;; From Norvig's PAIP, using the hash-pipe data structure above.

(defmacro defun-memo (fn args &body body)
  "Define a memoized function."
  `(memoize (defun ,fn ,args . ,body)))

(defun memo (fn &key (key #'first) (test #'eql) name depth)
  "Return a memo-function of fn."
  (let ((table (make-hash-pipe 
                  :hash (make-hash-table :test test) :depth depth)))
    (setf (get name 'memo) table)
    #'(lambda (&rest args)
        (dbg :memo "test key ~A args ~A" key args)
        (let ((k (funcall key args)))
          (dbg :memo "hashpipe key ~A" k)
          (multiple-value-bind (val found-p)
              (gethash-pipe k table)
            (dbg :memo "found-p ~A" found-p)
            (if found-p val
                (setf (gethash-pipe k table) (apply fn args))))))))

(defun memoize (fn-name &key (key #'first) (test #'eql) (depth 300))
  "Replace fn-name's global definition with a memoized version."
  (clear-memoize fn-name)
  (setf (symbol-function fn-name)
        (memo (symbol-function fn-name)
              :name fn-name :key key :test test :depth depth)))

(defun clear-memoize (fn-name)
  "Clear the hash table from a memo function."
  (let ((table (get fn-name 'memo)))
    (when table  
      (clrhash (hash-pipe-hash table))
      (setf (hash-pipe-keys table) nil) )))

;;;; Delayed computation:
;; From Norvig's PAIP.

(defstruct delay value (computed? nil))

(defmacro delay (&rest body)
  "A computation that can be executed later by FORCE."
  `(make-delay :value #'(lambda () . ,body)))

(defun force (delay)
  "Do a delayed computation, or fetch its previously-computed value."
  (if (delay-computed? delay)
      (delay-value delay)
      (prog1 (setf (delay-value delay) (funcall (delay-value delay)))
             (setf (delay-computed? delay) t))))
